"""
Support for accessing data in a repository with some "known locations".
"""
import re
import json
import pathlib
import functools
import webbrowser

import attr

from clldutils.misc import lazyproperty
from clldutils.path import git_describe
from clldutils.attrlib import valid_range
from clldutils.metadata import Metadata
from clldutils.jsonlib import load

VERSION_NUMBER_PATTERN = re.compile(
    r'v(?P<number>(?P<major>[0-9]+)\.(?P<minor>[0-9]+)(\.(?P<patch>[0-9]+))?)$')


#
# Common attributes of data objects
#
def latitude():
    return attr.ib(
        converter=lambda s: None if s is None or s == '' else float(s),
        validator=valid_range(-90, 90, nullable=True))


def longitude():
    return attr.ib(
        converter=lambda s: None if s is None or s == '' else float(s),
        validator=valid_range(-180, 180, nullable=True))


def value_ascsv(v):
    if v is None:
        return ''
    elif isinstance(v, float):
        return "{0:.5f}".format(v)
    elif isinstance(v, dict):
        return json.dumps(v)
    elif isinstance(v, list):
        return ';'.join(v)
    return "{0}".format(v)


@attr.s
class DataObject(object):

    @classmethod
    def fieldnames(cls):
        return [f.name for f in attr.fields(cls)]

    def ascsv(self):
        res = []
        for f, v in zip(attr.fields(self.__class__), attr.astuple(self)):
            res.append((f.metadata.get('ascsv') or value_ascsv)(v))
        return res


def assert_release(repos):
    match = VERSION_NUMBER_PATTERN.match(git_describe(repos))
    assert match, 'Repository is not checked out to a valid release tag'
    return match.group('number')  # pragma: no cover


class API(object):
    """An API base class to provide programmatic access to data in a git repository."""

    # A light-weight way to specifiy a default repository location (without having to
    # overwrite __init__)
    __repos_path__ = None
    __default_metadata__ = None

    def __init__(self, repos=None):
        self.repos = pathlib.Path(repos or self.__repos_path__)

    def __str__(self):
        name = self.repos.resolve().name if self.repos.exists() else self.repos.name
        return '<{0} repository {1} at {2}>'.format(
            name, git_describe(self.repos), self.repos)

    def path(self, *comps):
        return self.repos.joinpath(*comps)

    @lazyproperty
    def dataset_metadata(self):
        mdp = self.repos / 'metadata.json'
        return Metadata.from_jsonld(
            load(mdp) if mdp.exists() else {}, defaults=self.__default_metadata__)

    @property
    def appdir(self):
        return self.path('app')

    @property
    def appdatadir(self):
        return self.appdir.joinpath('data')

    @classmethod
    def app_wrapper(cls, func):
        @functools.wraps(func)
        def wrapper(args):
            if isinstance(args.repos, cls):
                api = args.repos
            else:
                api = cls(args.repos)
            if (not api.appdatadir.exists()) \
                    or getattr(args, 'recreate', False) \
                    or getattr(args, 'force', False) \
                    or (hasattr(args, 'args') and '--recreate' in args.args):
                api.appdatadir.mkdir(exist_ok=True)
                args.api = api
                func(args)
            index = api.appdir / 'index.html'
            if index.exists():
                webbrowser.open(index.resolve().as_uri())
        return wrapper

    def assert_release(self):
        return assert_release(self.repos)



import re
import itertools
import collections

ID_PATTERN = re.compile(r'^[a-zA-Z\-_0-9]+$')


class Source(collections.OrderedDict):
    """Bibliographic metadata about a source used for some analysis in a linguistic database.

    Following BibTeX-style, a Source is just an ordered list of key-value pairs, augmented
    with an id (a.k.a. BibTeX citekey) and a genre (a.k.a. Entry Types).

    .. note::

        We do restrict the allowed syntax for the id to make sure it can safely be used
        as path component in a URL. To skip this check, pass `_check_id=False` to the
        constructor.
    """

    def __init__(self, genre, id_, *args, **kw):
        if kw.pop('_check_id', True) and not ID_PATTERN.match(id_):
            raise ValueError(id_)
        self.genre = genre
        self.id = id_
        super(Source, self).__init__(*args, **kw)

    def __bool__(self):  # pragma: no cover
        return True

    __nonzero__ = __bool__

    def __str__(self):
        return self.text()

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.id)

    @classmethod
    def from_entry(cls, key, entry):
        """
        Factory method to initialize a `Source` instance from a `pybtex.database.Entry`.

        :param key: Citation key, e.g. a key in `pybtex.database.BibliographyData.entries`.
        :param entry: `pybtex.database.Entry`
        """
        kw = {k: v for k, v in entry.fields.items()}
        for role in (entry.persons or []):
            if entry.persons[role]:
                kw[role] = ' and '.join('%s' % p for p in entry.persons[role])
        return cls(entry.type, key, **kw)

    @classmethod
    def from_bibtex(cls, bibtexString, lowercase=False, _check_id=True):
        source = None

        # the following patterns are designed to match preprocessed input lines.
        # i.e. the configuration values given in the bibtool resource file used to
        # generate the bib-file have to correspond to these patterns.

        # in particular, we assume all key-value-pairs to fit on one line,
        # because we don't want to deal with nested curly braces!
        lines = bibtexString.strip().split('\n')

        # genre and key are parsed from the @-line:
        atLine = re.compile(r"^@(?P<genre>[a-zA-Z_]+)\s*{\s*(?P<key>[^,]*)\s*,\s*")

        # since all key-value pairs fit on one line, it's easy to determine the
        # end of the value: right before the last closing brace!
        fieldLine = re.compile(r'\s*(?P<field>[a-zA-Z_]+)\s*=\s*(\{|")(?P<value>.+)')

        endLine = re.compile(r"}\s*")

        while lines:
            line = lines.pop(0)
            if not source:
                m = atLine.match(line)
                if m:
                    source = cls(
                        m.group('genre').strip().lower(),
                        m.group('key').strip(),
                        _check_id=_check_id)
            else:
                m = fieldLine.match(line)
                if m:
                    value = m.group('value').strip()
                    if value.endswith(','):
                        value = value[:-1].strip()
                    if value.endswith('}') or value.endswith('"'):
                        field = m.group('field')
                        if lowercase:
                            field = field.lower()
                        source[field] = value[:-1].strip()
                else:
                    m = endLine.match(line)
                    if m:
                        break
                    # Note: fields with names not matching the expected pattern are simply
                    # ignored.

        return source

    def bibtex(self):
        """Represent the source in BibTeX format.

        :return: string encoding the source in BibTeX syntax.
        """
        m = max(itertools.chain(map(len, self), [0]))
        fields = ("  %s = {%s}" % (k.ljust(m), self[k]) for k in self)
        return "@%s{%s,\n%s\n}" % (
            getattr(self.genre, 'value', self.genre), self.id, ",\n".join(fields))

    _genre_note = {
        'phdthesis': 'Doctoral dissertation',
        'mastersthesis': 'MA thesis',
        'unpublished': 'unpublished',
    }

    def get_with_translation(self, key):
        res = self.get(key)
        if res and self.get(key + '_english'):
            res = '{0} [{1}]'.format(res, self.get(key + '_english'))
        return res

    def text(self):
        """Linearize the bib source according to the rules of the unified style.

        Book:
        author. year. booktitle. (series, volume.) address: publisher.

        Article:
        author. year. title. journal volume(issue). pages.

        Incollection:
        author. year. title. In editor (ed.), booktitle, pages. address: publisher.

        .. seealso::

            http://celxj.org/downloads/UnifiedStyleSheet.pdf
            https://github.com/citation-style-language/styles/blob/master/\
            unified-style-linguistics.csl
        """
        def fmt_edition(e):
            try:
                e = int(e)
                return "%d%s" % (e, "tsnrhtdd"[(e // 10 % 10 != 1) * (e % 10 < 4) * e % 10::4])
            except ValueError:  # pragma: no cover
                return e

        genre = getattr(self.genre, 'value', self.genre)
        pages_at_end = genre in (
            'book',
            'phdthesis',
            'mastersthesis',
            'misc',
            'techreport')
        thesis = genre in ('phdthesis', 'mastersthesis')

        if self.get('editor'):
            editors = self['editor']
            affix = 'eds' if ' and ' in editors or '&' in editors else 'ed'
            editors = " %s (%s.)" % (editors, affix)
        else:
            editors = None

        res = [self.get('author', editors), self.get('year', 'n.d')]
        if genre == 'book':
            res.append(
                self.get_with_translation('booktitle') or  # noqa: W504
                self.get_with_translation('title'))
            series = ', '.join(filter(None, [self.get('series'), self.get('volume')]))
            if series:
                res.append('(%s.)' % series)
        elif genre == 'misc':
            # in case of misc records, we use the note field in case a title is missing.
            res.append(self.get_with_translation('title') or self.get('note'))
        else:
            res.append(self.get_with_translation('title'))

        if genre == 'article':
            atom = ' '.join(filter(None, [self.get('journal'), self.get('volume')]))
            if self.get('issue'):
                atom += '(%s)' % self['issue']
            res.append(atom)
            res.append(self.get('pages'))
        elif genre == 'incollection' or genre == 'inproceedings':
            prefix = 'In'
            atom = ''
            if editors:
                atom += editors
            if self.get('booktitle'):
                if atom:
                    atom += ','
                atom += " %s" % self.get_with_translation('booktitle')
            if self.get('pages'):
                atom += ", %s" % self['pages']
            res.append(prefix + atom)
        else:
            # check for author to make sure we haven't included the editors yet.
            if editors and self.get('author'):
                res.append("In %s" % editors)

            for attr in [
                'journal',
                'volume' if genre != 'book' else None,
            ]:
                if attr and self.get(attr):
                    res.append(self.get(attr))

            if self.get('issue'):
                res.append("(%s)" % self['issue'])

            if not pages_at_end and self.get('pages'):
                res.append(self['pages'])

        if self.get('publisher'):
            if self.get('edition'):
                res.append('{} edn'.format(fmt_edition(self.get('edition'))))
            publisher = self.get('publisher')
            if self.get('address') and publisher.startswith('{}:'.format(self['address'])):
                res.append(self['publisher'])
            else:
                res.append(": ".join(filter(None, [self.get('address'), self['publisher']])))
        else:
            if genre == 'misc' and self.get('howpublished'):
                res.append(self.get('howpublished'))

        if not thesis and pages_at_end and self.get('pages'):
            res.append(self['pages'] + 'pp')

        note = self.get('note') or self._genre_note.get(genre)
        if note and note not in res:
            if thesis:
                joiner = ','
                if self.get('school'):
                    note += '{0} {1}'.format(joiner, self.get('school'))
                    joiner = ';'
                if self.get('pages'):
                    note += '{0} {1}pp.'.format(joiner, self.get('pages'))
            res.append('(%s)' % note)

        return ' '.join(
            x if x.endswith(('.', '.)')) else '%s.' % x for x in res if x)



"""
Functionality to manage colors for visualizations.
"""
import colorsys
import itertools
import fractions
import math

__all__ = [
    'brightness',
    'is_bright',
    'rgb_as_hex',
    'diverging_colors',
    'qualitative_colors',
    'sequential_colors'
]


def _to_rgb(s):
    def f2i(d):
        assert 0 <= d <= 1
        res = int(math.floor(d * 256))
        if res == 256:
            res = 255
        return res

    if isinstance(s, (tuple, list)):
        assert len(s) == 3
        if isinstance(s[0], (float, fractions.Fraction)):
            s = [f2i(d) for d in s]
        return s
    assert isinstance(s, str)
    if s.startswith('#'):
        s = s[1:]
    if len(s) == 3:
        s = ''.join(c + c for c in s)
    assert len(s) == 6
    return tuple(int(c, 16) for c in [s[i:i + 2] for i in range(0, 6, 2)])


def rgb_as_hex(s):
    return '#{0:02X}{1:02X}{2:02X}'.format(*_to_rgb(s))


def brightness(color):
    """
    see https://www.w3.org/TR/AERT/#color-contrast
    """
    R, G, B = _to_rgb(color)
    return 0.299 * R + 0.587 * G + 0.114 * B


def is_bright(color):
    # 125 seems to be a common cut-off above which to regard a color as "bright".
    return brightness(color) > 125


def qualitative_colors(n, set=None):
    """
    Choses `n` distinct colors suitable for visualizing categorical variables.

    ..seealso:: https://en.wikipedia.org/wiki/Categorical_variable

    :param n: number of distinct colors to return
    :param set: name of a particular color set to choose from
    :return: list of `n` hex color codes
    """
    if n <= 11 and set == 'boynton':
        # R. M. Boynton. Eleven colors that are almost never confused.
        # In B. E. Rogowitz, editor,
        # Proceedings of the SPIE Symposium: Human Vision, Visual Processing, and Digital
        # Display, volume 1077, pages 322{332, Bellingham, WA, 1989.
        # SPIE Int. Soc. Optical Engineering.
        return [
            rgb_as_hex(c) for c in
            [
                (91, 0, 13),
                (0, 255, 223),
                (23, 169, 255),
                (255, 232, 0),
                (8, 0, 91),
                (255, 208, 198),
                (4, 255, 4),
                (0, 0, 255),
                (0, 79, 0),
                (255, 21, 205),
                (255, 0, 0),
            ][:n]]
    if n <= 12 and set == 'tol':
        # https://personal.sron.nl/~pault/colourschemes.pdf
        # as implemented by drmccloy here https://github.com/drammock/colorblind
        cols = ['#4477AA', '#332288', '#6699CC', '#88CCEE', '#44AA99', '#117733',
                '#999933', '#DDCC77', '#661100', '#CC6677', '#AA4466', '#882255',
                '#AA4499']
        indices = [[0],
                   [0, 9],
                   [0, 7, 9],
                   [0, 5, 7, 9],
                   [1, 3, 5, 7, 9],
                   [1, 3, 5, 7, 9, 12],
                   [1, 3, 4, 5, 7, 9, 12],
                   [1, 3, 4, 5, 6, 7, 9, 12],
                   [1, 3, 4, 5, 6, 7, 9, 11, 12],
                   [1, 3, 4, 5, 6, 7, 8, 9, 11, 12],
                   [1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 12],
                   [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]]
        return [cols[ix] for ix in indices[n - 1]]

    if n <= 22:
        # theory:
        # https://eleanormaclure.files.wordpress.com/2011/03/colour-coding.pdf (page 5)
        # kelly's colors:
        #  https://i.kinja-img.com/gawker-media/image/upload/1015680494325093012.JPG
        return [
            rgb_as_hex(c) for c in
            [
                'F2F3F4',
                '222222',
                'F3C300',
                '875692',
                'F38400',
                'A1CAF1',
                'BE0032',
                'C2B280',
                '848482',
                '008856',
                'E68FAC',
                '0067A5',
                'F99379',
                '604E97',
                'F6A600',
                'B3446C',
                'DCD300',
                '882D17',
                '8DB600',
                '654522',
                'E25822',
                '2B3D26'][:n]
        ]

    #
    # taken from https://stackoverflow.com/a/13781114
    #
    def zenos_dichotomy():
        """
        http://en.wikipedia.org/wiki/1/2_%2B_1/4_%2B_1/8_%2B_1/16_%2B_%C2%B7_%C2%B7_%C2%B7
        """
        for k in itertools.count():
            yield fractions.Fraction(1, 2**k)

    def getfracs():
        yield 0
        for k in zenos_dichotomy():
            i = k.denominator  # [1,2,4,8,16,...]
            for j in range(1, i, 2):
                yield fractions.Fraction(j, i)

    def genhsv(h):
        for s in [fractions.Fraction(6, 10)]:  # optionally use range
            for v in [fractions.Fraction(8, 10), fractions.Fraction(5, 10)]:  # could use range too
                yield (h, s, v)  # use bias for v here if you use range

    def gethsvs():
        return itertools.chain.from_iterable(map(genhsv, getfracs()))

    return [
        rgb_as_hex(c) for c in
        itertools.islice((colorsys.hsv_to_rgb(*x) for x in gethsvs()), n)]


def sequential_colors(n):
    """
    Between 3 and 9 sequential colors.
    """
    # https://personal.sron.nl/~pault/
    # as implemented by drmccloy here https://github.com/drammock/colorblind
    assert 3 <= n <= 9
    cols = ['#FFFFE5', '#FFFBD5', '#FFF7BC', '#FEE391', '#FED98E', '#FEC44F',
            '#FB9A29', '#EC7014', '#D95F0E', '#CC4C02', '#993404', '#8C2D04',
            '#662506']
    indices = [[2, 5, 8],
               [1, 3, 6, 9],
               [1, 3, 6, 8, 10],
               [1, 3, 5, 6, 8, 10],
               [1, 3, 5, 6, 7, 9, 10],
               [0, 2, 3, 5, 6, 7, 9, 10],
               [0, 2, 3, 5, 6, 7, 9, 10, 12]]
    return [cols[ix] for ix in indices[n - 3]]


def diverging_colors(n):
    """
    Between 3 and 11 diverging colors
    """
    # https://personal.sron.nl/~pault/
    # as implemented by drmccloy here https://github.com/drammock/colorblind
    assert 3 <= n <= 11
    cols = ['#3D52A1', '#3A89C9', '#008BCE', '#77B7E5', '#99C7EC', '#B4DDF7',
            '#E6F5FE', '#FFFAD2', '#FFE3AA', '#F9BD7E', '#F5A275', '#ED875E',
            '#D03232', '#D24D3E', '#AE1C3E']
    indices = [[4, 7, 10],
               [2, 5, 9, 12],
               [2, 5, 7, 9, 12],
               [1, 4, 6, 8, 10, 13],
               [1, 4, 6, 7, 8, 10, 13],
               [1, 3, 5, 6, 8, 9, 11, 13],
               [1, 3, 5, 6, 7, 8, 9, 11, 13],
               [0, 1, 3, 5, 6, 8, 9, 11, 13, 14],
               [0, 1, 3, 5, 6, 7, 8, 9, 11, 13, 14]]
    return [cols[ix] for ix in indices[n - 3]]



"""Generic utility functions."""

import re
import base64
import string
import pathlib
import warnings
import mimetypes
import unicodedata

from six import PY3, string_types, text_type, binary_type, iteritems

__all__ = [
    'data_url', 'log_or_raise', 'nfilter', 'to_binary', 'dict_merged', 'NoDefault', 'NO_DEFAULT',
    'xmlchars', 'format_size', 'UnicodeMixin', 'slug', 'encoded', 'lazyproperty',
]


def deprecated(msg):
    warnings.simplefilter('always', DeprecationWarning)
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
    warnings.simplefilter('default', DeprecationWarning)


def data_url(content, mimetype=None):
    """
    Returns content encoded as base64 Data URI.

    :param content: bytes or str or Path
    :param mimetype: mimetype for
    :return: str object (consisting only of ASCII, though)

    .. seealso:: https://en.wikipedia.org/wiki/Data_URI_scheme
    """
    if isinstance(content, pathlib.Path):
        if not mimetype:
            mimetype = mimetypes.guess_type(content.name)[0]
        with content.open('rb') as fp:
            content = fp.read()
    else:
        if isinstance(content, text_type):
            content = content.encode('utf8')
    return "data:{0};base64,{1}".format(
        mimetype or 'application/octet-stream', base64.b64encode(content).decode())


def log_or_raise(msg, log=None, level='warning', exception_cls=ValueError):
    if log:
        getattr(log, level)(msg)
    else:
        raise exception_cls(msg)


def nfilter(seq):
    """Replacement for python 2's filter(None, seq).

    :return: a list filtered from seq containing only truthy items.
    """
    return [e for e in seq if e]


def to_binary(s, encoding='utf8'):
    """Portable cast function.

    In python 2 the ``str`` function which is used to coerce objects to bytes does not
    accept an encoding argument, whereas python 3's ``bytes`` function requires one.

    :param s: object to be converted to binary_type
    :return: binary_type instance, representing s.
    """
    if PY3:  # pragma: no cover
        return s if isinstance(s, binary_type) else binary_type(s, encoding=encoding)
    return binary_type(s)  # pragma: no cover


def dict_merged(d, _filter=None, **kw):
    """Update dictionary d with the items passed as kw if the value passes _filter."""
    def f(s):
        if _filter:
            return _filter(s)
        return s is not None
    d = d or {}
    for k, v in iteritems(kw):
        if f(v):
            d[k] = v
    return d


class NoDefault(object):

    def __repr__(self):
        return '<NoDefault>'


#: A singleton which can be used to distinguish no-argument-passed from None passed as
#: argument in callables with optional arguments.
NO_DEFAULT = NoDefault()


def xmlchars(text):
    """Not all of UTF-8 is considered valid character data in XML ...

    Thus, this function can be used to remove illegal characters from ``text``.
    """
    invalid = list(range(0x9))
    invalid.extend([0xb, 0xc])
    invalid.extend(range(0xe, 0x20))
    return re.sub('|'.join('\\x%0.2X' % i for i in invalid), '', text)


def format_size(num):
    """Format byte-sizes.

    .. seealso:: http://stackoverflow.com/a/1094933
    """
    for x in ['bytes', 'KB', 'MB', 'GB']:
        if num < 1024.0 and num > -1024.0:
            return "%3.1f%s" % (num, x)
        num /= 1024.0
    return "%3.1f%s" % (num, 'TB')


class UnicodeMixin(object):
    """Portable label mixin."""

    def __unicode__(self):
        """a human readable label for the object."""
        return '%s' % self  # pragma: no cover

    def __str__(self):
        """a human readable label for the object, appropriately encoded (or not)."""
        deprecated("Use of deprecated class UnicodeMixin! Use object instead.")
        return self.__unicode__()


def slug(s, remove_whitespace=True, lowercase=True):
    """Condensed version of s, containing only lowercase alphanumeric characters."""
    res = ''.join(c for c in unicodedata.normalize('NFD', s)
                  if unicodedata.category(c) != 'Mn')
    if lowercase:
        res = res.lower()
    for c in string.punctuation:
        res = res.replace(c, '')
    res = re.sub(r'\s+', '' if remove_whitespace else ' ', res)
    res = res.encode('ascii', 'ignore').decode('ascii')
    assert re.match('[ A-Za-z0-9]*$', res)
    return res


def encoded(string, encoding='utf-8'):
    """Cast string to binary_type.

    :param string: six.binary_type or six.text_type
    :param encoding: encoding which the object is forced to
    :return: six.binary_type
    """
    assert isinstance(string, string_types) or isinstance(string, binary_type)
    if isinstance(string, text_type):
        return string.encode(encoding)
    try:
        # make sure the string can be decoded in the specified encoding ...
        string.decode(encoding)
        return string
    except UnicodeDecodeError:
        # ... if not use latin1 as best guess to decode the string before encoding as
        # specified.
        return string.decode('latin1').encode(encoding)


class lazyproperty(object):
    """Non-data descriptor caching the computed result as instance attribute.

    >>> class Spam(object):
    ...     @lazyproperty
    ...     def eggs(self):
    ...         return 'spamspamspam'
    >>> spam=Spam(); spam.eggs
    'spamspamspam'
    >>> spam.eggs='eggseggseggs'; spam.eggs
    'eggseggseggs'
    >>> Spam().eggs
    'spamspamspam'
    >>> Spam.eggs  # doctest: +ELLIPSIS
    <...lazyproperty object at 0x...>
    """

    def __init__(self, fget):
        self.fget = fget
        for attr in ('__module__', '__name__', '__doc__'):
            setattr(self, attr, getattr(fget, attr))

    def __get__(self, instance, owner):
        if instance is None:
            return self
        result = instance.__dict__[self.__name__] = self.fget(instance)
        return result



import re
import sys

from tabulate import tabulate

__all__ = ['Table', 'iter_markdown_tables', 'iter_markdown_sections']


class Table(list):

    def __init__(self, *cols, **kw):
        self.columns = list(cols)
        super(Table, self).__init__(kw.pop('rows', []))
        self._file = kw.pop('file', sys.stdout)
        self._kw = kw

    def render(self, sortkey=None, condensed=True, verbose=False, reverse=False, **kw):
        """

        :param sortkey: A callable which can be used as key when sorting the rows.
        :param condensed: Flag signalling whether whitespace padding should be collapsed.
        :param verbose: Flag signalling whether to output additional info.
        :param reverse: Flag signalling whether we should sort in reverse order.
        :param kw: Additional keyword arguments are passed to the `tabulate` function.
        :return: String representation of the table in the chosen format.
        """
        tab_kw = dict(tablefmt='pipe', headers=self.columns, floatfmt='.2f')
        tab_kw.update(self._kw)
        tab_kw.update(kw)
        res = tabulate(
            sorted(self, key=sortkey, reverse=reverse) if sortkey else self, **tab_kw)
        if tab_kw['tablefmt'] == 'pipe':
            if condensed:
                # remove whitespace padding around column content:
                res = re.sub(r'\|[ ]+', '| ', res)
                res = re.sub(r'[ ]+\|', ' |', res)
            if verbose:
                res += '\n\n(%s rows)\n\n' % len(self)
        return res

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        print(self.render(), file=self._file)


def iter_markdown_tables(text):
    """
    Parse tables from a markdown formatted text.

    :param text: `str` of markdown formatted text.
    :return: generator of (header, rows) pairs, where "header" is a `list` of column names and \
    rows is a `list` of `list`s of row values.
    """
    def split_row(line, outer_pipes):
        line = line.strip()
        if outer_pipes:
            assert line.startswith('|') and line.endswith('|'), 'inconsistent table formatting'
            line = line[1:-1].strip()
        return [c.strip() for c in line.split('|')]

    for header, rows, outer_pipes in _iter_table_blocks(text.splitlines()):
        yield split_row(header, outer_pipes), [split_row(row, outer_pipes) for row in rows]


def _iter_table_blocks(lines):
    # Tables are detected by
    # 1. A header line, i.e. a line with at least one `|`
    # 2. A line separating header and body of the form below
    SEP = re.compile(r'\s*\|?\s*:?--(-)+:?\s*(\|\s*:?--(-)+:?\s*)+\|?\s*')

    lines = list(lines)
    header, table, outer_pipes = None, [], False
    for i, line in enumerate(lines):
        if header:
            if '|' not in line:
                if table:
                    yield header, table, outer_pipes
                header, table, outer_pipes = None, [], False
            else:
                if not SEP.fullmatch(line):
                    table.append(line)
        else:
            if '|' in line and len(lines) > i + 1 and SEP.fullmatch(lines[i + 1]):
                header = line
                outer_pipes = lines[i + 1].strip().startswith('|')
    if table:
        yield header, table, outer_pipes


def iter_markdown_sections(text):
    """
    Parse sections from a markdown formatted text.

    Note: We only recognize the "#" syntax for marking section headings.

    :param text: `str` of markdown formatted text.
    :return: generator of (level, header, content) pairs, where "level" is an `int`, \
    "header" is the exact section heading (including "#"s and newline) or `None` and \
    "content" the markdown text of the section.
    """
    section_pattern = re.compile(r'(?P<level>[#]+)')
    lines, header, level = [], None, None
    for line in text.splitlines(keepends=True):
        match = section_pattern.match(line)
        if match:
            if lines:
                yield level, header, ''.join(lines)
            lines, header, level = [], line, len(match.group('level'))
        else:
            lines.append(line)
    if lines or header:
        yield level, header, ''.join(lines)



"""Programmatic access to the information of the ISO-639-3 standard

ISO-639-3 data is not distributed with this package, but we fetch the download listed at
http://www-01.sil.org/iso639-3/download.asp
"""

import re
import string
import datetime
import functools
import collections
import urllib.request

from csvw.dsv import iterrows

from clldutils.path import TemporaryDirectory, Path
from clldutils.ziparchive import ZipArchive

BASE_URL = "https://iso639-3.sil.org/"
ZIP_NAME_PATTERN = re.compile(
    r'(?P<name>sites/iso639-3/files/downloads/iso-639-3_Code_Tables_[0-9]{8}.zip)"')
TABLE_NAME_PATTERN = re.compile(r'/iso-639-3(?P<name_and_date>[^.]*)\.tab')
DATESTAMP_PATTERN = re.compile(r'(2[0-9]{3})([0-1][0-9])([0-3][0-9])')
USER_AGENT = 'Mozilla'  # It seems a python user-agent doesn't cut it anymore.

# For some reason, the retirements code table gives the wrong replacement codes in two
# cases (although they are described correctly on the website):
CHANGE_TO_ERRATA = {
    'guv': ['duz'],
    'ymt': ['mtm'],
}


def _open(path):
    return urllib.request.urlopen(
        urllib.request.Request(BASE_URL + path, headers={'User-Agent': USER_AGENT}))


class Table(list):

    def __init__(self, name_and_date, date, fp):
        parts = name_and_date.split('_')
        # The ISO 639-3 code tables from 2020-05-15 contain a table with a
        # malformed name - having an excess "0" in the date stamp.
        if parts[-1] == '202000515':  # pragma: no cover
            date = '20200515'
        digits = map(int, DATESTAMP_PATTERN.match(date).groups())
        self.date = datetime.date(*digits)
        name = '_'.join([p for p in parts if not DATESTAMP_PATTERN.match(p)])
        if name.startswith(('_', '-')):
            name = name[1:]
        if not name:
            name = 'Codes'
        self.name = name
        super(Table, self).__init__(iterrows(
            [line for line in fp.splitlines() if line.strip()],  # strip malformed lines.
            dicts=True,
            delimiter='\t'))


def download_tables(outdir=None):
    match = ZIP_NAME_PATTERN.search(_open('code_tables/download_tables').read().decode('utf-8-sig'))
    if not match:
        raise ValueError('no matching zip file name found')  # pragma: no cover
    target = Path(outdir or '.').joinpath(match.group('name').split('/')[-1])
    with target.open('wb') as fp:
        fp.write(_open(match.group('name')).read())
    return target


def iter_tables(zippath=None):
    with TemporaryDirectory() as tmp:
        if not zippath:
            zippath = download_tables(tmp)

        with ZipArchive(zippath) as archive:
            for name in archive.namelist():
                date = DATESTAMP_PATTERN.search(name)
                date = name[date.start():date.end()]
                match = TABLE_NAME_PATTERN.search(name)
                if match:
                    yield Table(match.group('name_and_date'), date, archive.read_text(name))


@functools.total_ordering
class Code(object):

    _code_pattern = re.compile(r'\[([a-z]{3})]')
    _scope_map = {
        'I': 'Individual',
        'M': 'Macrolanguage',
        'S': 'Special',
    }
    _type_map = {
        'L': 'Living',
        'E': 'Extinct',
        'A': 'Ancient',
        'H': 'Historical',
        'C': 'Constructed',
        'S': 'Special',
    }
    _rtype_map = {
        'C': 'change',
        'D': 'duplicate',
        'N': 'non-existent',
        'S': 'split',
        'M': 'merge',
    }

    def __init__(self, item, tablename, registry):
        code = item['Id']
        self._change_to = []
        self.retired = False
        if tablename == 'Codes':
            self._scope = self._scope_map[item['Scope']]
            self._type = self._type_map[item['Language_Type']]
        elif tablename == 'Retirements':
            self._scope = 'Retirement'
            self._type = self._rtype_map[item['Ret_Reason']] if item['Ret_Reason'] else None
            self.retired = datetime.date(*map(int, item['Effective'].split('-')))
            if code in CHANGE_TO_ERRATA:
                self._change_to = CHANGE_TO_ERRATA[code]  # pragma: no cover
            else:
                if item['Change_To']:
                    assert item['Change_To'] != code
                    self._change_to = [item['Change_To']]
                elif item['Ret_Remedy']:
                    self._change_to = [
                        c for c in self._code_pattern.findall(item['Ret_Remedy'])
                        if c != code]
        elif tablename == 'Local':
            self._scope = 'Local'
            self._type = 'Special'
        else:
            raise ValueError(tablename)  # pragma: no cover

        self.code = code
        self.name = item['Ref_Name']
        self._registry = registry

    @property
    def type(self):
        return '{0}/{1}'.format(self._scope, self._type)

    @property
    def is_retired(self):
        return bool(self.retired)

    @property
    def change_to(self):
        res = []
        for code in self._change_to:
            code = self._registry[code]
            if not code.is_retired:
                res.append(code)
            else:
                res.extend(code.change_to)
        return res

    @property
    def is_local(self):
        return self._scope == 'Local'

    @property
    def is_macrolanguage(self):
        return self._scope == 'Macrolanguage'

    @property
    def extension(self):
        if self.is_macrolanguage:
            return [self._registry[c] for c in self._registry._macrolanguage[self.code]]
        return []

    def __hash__(self):
        return hash(self.code)

    def __eq__(self, other):
        return self.code == other.code

    def __lt__(self, other):
        return self.code < other.code

    def __repr__(self):
        return '<ISO-639-3 [{0}] {1}>'.format(self.code, self.type)

    def __str__(self):
        return '{0} [{1}]'.format(self.name, self.code)


class ISO(collections.OrderedDict):

    def __init__(self, zippath=None):
        self._tables = {t.name: t for t in iter_tables(zippath=zippath)}
        if zippath and DATESTAMP_PATTERN.search(zippath.name):
            digits = map(int, DATESTAMP_PATTERN.search(zippath.name).groups())
            self.date = datetime.date(*digits)
        else:
            self.date = max(t.date for t in self._tables.values())
        self._macrolanguage = collections.defaultdict(list)
        for item in self._tables['macrolanguages']:
            self._macrolanguage[item['M_Id']].append(item['I_Id'])
        super(ISO, self).__init__()
        for tablename in ['Codes', 'Retirements']:
            for item in self._tables[tablename]:
                if item['Id'] not in self:
                    # Note: we don't keep historical retirements, i.e. ones that have only
                    # been in effect for some time. E.g. lcq has been changed to ppr
                    # from 2012-02-03 until 2013-01-23 when it was changed back to lcq
                    self[item['Id']] = Code(item, tablename, self)
        for code in ['q' + x + y
                     for x in string.ascii_lowercase[:string.ascii_lowercase.index('t') + 1]
                     for y in string.ascii_lowercase]:
            self[code] = Code(dict(Id=code, Ref_Name=None), 'Local', self)

    def __str__(self):
        return 'ISO 639-3 code tables from {0}'.format(self.date)

    def by_type(self, type_):
        return [c for c in self.values() if c._type == type_]

    @property
    def living(self):
        return self.by_type('Living')

    @property
    def extinct(self):
        return self.by_type('Extinct')

    @property
    def ancient(self):
        return self.by_type('Ancient')

    @property
    def historical(self):
        return self.by_type('Historical')

    @property
    def constructed(self):
        return self.by_type('Constructed')

    @property
    def special(self):
        return self.by_type('Special')

    @property
    def retirements(self):
        return [c for c in self.values() if c.is_retired]

    @property
    def macrolanguages(self):
        return [c for c in self.values() if c.is_macrolanguage]

    @property
    def languages(self):
        return [c for c in self.values()
                if not c.is_macrolanguage and not c.is_retired and not c.is_local]



import re
import math

__all__ = ['Coordinates', 'dec2degminsec', 'degminsec2dec', 'degminsec']

DEGREES = "°"
MINUTES = "\u2032"
SECONDS = "\u2033"

PATTERNS = {
    'lat_alnum': re.compile(r"(?P<deg>\d+)d(?P<min>[0-9]+)?(?P<sec>'\d+'')?(?P<hem>S|N)"),
    'lon_alnum': re.compile(r"(?P<deg>\d+)d(?P<min>\d+)?(?P<sec>'\d+'')?(?P<hem>E|W)"),
    'lat_degminsec': re.compile(
        r'(?P<deg>\d+)\s*%s\s*((?P<min>\d+)\s*%s\s*)?((?P<sec>[\d.]+)\s*%s\s*)?(?P<hem>S|N)' % (
            DEGREES, MINUTES, SECONDS)),
    'lon_degminsec': re.compile(
        r'(?P<deg>\d+)\s*%s\s*((?P<min>\d+)\s*%s\s*)?((?P<sec>[\d.]+)\s*%s\s*)?(?P<hem>E|W)' % (
            DEGREES, MINUTES, SECONDS)),
}


def degminsec(dec, hemispheres):
    """
    >>> degminsec(2.4, 'NS')
    "2°24'N"
    """
    if 'N' in hemispheres:
        return Coordinates(dec, 0).lat_to_string(format='ascii')
    return Coordinates(0, dec).lon_to_string(format='ascii')


def dec2degminsec(dec):
    """
    convert a floating point number of degrees to a triple
    (int degrees, int minutes, float seconds)

    >>> assert dec2degminsec(30.50) == (30, 30, 0.0)
    """
    degrees = int(math.floor(dec))
    dec = (dec - int(math.floor(dec))) * 60
    minutes = int(math.floor(dec))
    dec = (dec - int(math.floor(dec))) * 60
    seconds = dec
    return degrees, minutes, seconds


def degminsec2dec(degrees, minutes, seconds):
    """
    convert a triple (int degrees, int minutes, float seconds) to
    a floating point number of degrees

    >>> assert dec2degminsec(degminsec2dec(30,30,0.0)) == (30,30,0.0)
    """
    dec = float(degrees)
    if minutes:
        dec += float(minutes) / 60
    if seconds:
        dec += float(seconds) / 3600
    return dec


class Coordinates(object):
    """
    >>> c = Coordinates('13dN', 0)
    >>> assert c.latitude >= 13
    >>> assert c.latitude <= 13.1
    >>> c = Coordinates(0, 0)
    >>> assert c.lat_to_string() == '0dN'
    >>> assert c.lon_to_string() == '0dE'
    >>> c = Coordinates(12.17, 92.83)
    >>> assert c.lat_to_string() == '12d10N'
    >>> assert c.lon_to_string() == '92d49E'
    >>> c = Coordinates(-12.17, -92.83)
    >>> assert c.lat_to_string() == '12d10S'
    >>> assert c.lon_to_string() == '92d49W'
    >>> lat, lon = '12d30N', '60d30E'
    >>> c = Coordinates(lat, lon)
    >>> assert c.lat_to_string() == lat
    >>> assert c.lon_to_string() == lon
    """

    def __init__(self, lat, lon, format='alnum'):
        if isinstance(lat, float):
            self.latitude = lat
        elif isinstance(lat, int):
            self.latitude = float(lat)
        else:
            self.latitude = self.lat_from_string(lat, format)

        if isinstance(lon, float):
            self.longitude = lon
        elif isinstance(lon, int):
            self.longitude = float(lon)
        else:
            self.longitude = self.lon_from_string(lon, format)

    def _match(self, string, type, format):
        if isinstance(string, bytes):
            string = string.decode('utf8')

        if type + '_' + format in PATTERNS:
            p = PATTERNS[type + '_' + format]
        else:
            p = PATTERNS[type + '_alnum']

        m = p.match(string)
        if not m:
            raise ValueError(string)
        return m

    def lat_from_string(self, lat, format='alnum'):
        m = self._match(lat, 'lat', format)
        dec = degminsec2dec(m.group('deg'), m.group('min'), m.group('sec'))
        if m.group('hem') == 'S':
            dec = -dec
        return dec

    def lon_from_string(self, lon, format='alnum'):
        m = self._match(lon, 'lon', format)
        dec = degminsec2dec(m.group('deg'), m.group('min'), m.group('sec'))
        if m.group('hem') == 'W':
            dec = -dec
        return dec

    def _format(self, degrees, minutes, seconds, hemisphere, format):
        seconds = int(round(seconds))
        if seconds == 60:
            minutes += 1
            seconds = 0

        if 120 > minutes >= 60:  # pragma: no cover
            # This case cannot really happen, because we only ever feed the results of
            # dec2degminsec into this method.
            degrees += 1
            minutes -= 60

        if format == 'alnum':
            res = "%sd" % degrees
            if minutes:
                res += "%02d" % minutes
            res += hemisphere
            return res

        if format == 'ascii':
            res = "%s°" % degrees
            if minutes:
                res += "{0:0>2d}'".format(minutes)
            if seconds:
                res += '{0:0>2f}"'.format(seconds)
            res += hemisphere
            return res

        res = "%s%s" % (degrees, DEGREES)

        if minutes:
            res += " %s%s" % (minutes, MINUTES)

        if seconds:
            res += " %s%s" % (seconds, SECONDS)
        res += " %s" % hemisphere
        return res

    def lat_to_string(self, format='alnum'):
        if self.latitude < 0:
            hemisphere = 'S'
        else:
            hemisphere = 'N'
        degrees, minutes, seconds = dec2degminsec(abs(self.latitude))
        return self._format(degrees, minutes, seconds, hemisphere, format)

    def lon_to_string(self, format='alnum'):
        if self.longitude < 0:
            hemisphere = 'W'
        else:
            hemisphere = 'E'
        degrees, minutes, seconds = dec2degminsec(abs(self.longitude))
        return self._format(degrees, minutes, seconds, hemisphere, format)



import io
import zipfile


class ZipArchive(zipfile.ZipFile):

    _init_defaults = {
        'compression': zipfile.ZIP_DEFLATED,
        'allowZip64': True,
    }

    def __init__(self, fname, mode='r', **kwargs):
        for k, v in self._init_defaults.items():
            kwargs.setdefault(k, v)
        super(ZipArchive, self).__init__(str(fname), mode=mode, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def read_text(self, name, encoding='utf-8-sig'):
        if name in self.namelist():
            return io.TextIOWrapper(self.open(name), encoding=encoding).read()

    def write_text(self, text, name, _encoding='utf-8'):
        if not isinstance(text, bytes):
            text = text.encode(_encoding)
        self.writestr(name, text)



import io
import re
import pathlib
import configparser


class INI(configparser.ConfigParser):

    @staticmethod
    def format_list(items):
        return ''.join('\n' + item for item in items)

    @classmethod
    def from_file(cls, fname, encoding='utf-8', **kw):
        obj = cls(**kw)
        obj.read(str(fname), encoding=encoding)
        return obj

    def write_string(self, **kw):
        res = io.StringIO()
        res.write('# -*- coding: utf-8 -*-\n')
        super(INI, self).write(res, **kw)
        return res.getvalue()

    def set(self, section, option, value=None):
        if value is None:
            return
        if not self.has_section(section):
            self.add_section(section)
        if isinstance(value, (list, tuple)):
            value = self.format_list(value)
        elif not isinstance(value, str):
            value = '%s' % value
        super(INI, self).set(section, option, value)

    def getlist(self, section, option):
        return self.get(section, option, fallback='').strip().splitlines()

    def gettext(self, section, option, whitespace_preserving_prefix='.'):
        """
        While configparser supports multiline values, it does this at the expense of
        stripping leading whitespace for each line in such a value. Sometimes we want
        to preserve such whitespace, e.g. to be able to put markdown with nested lists
        into INI files. We support this be introducing a special prefix, which is
        prepended to lines starting with whitespace in `settext` and stripped in
        `gettext`.
        """
        lines = []
        for line in self.get(section, option, fallback='').splitlines():
            if re.match(re.escape(whitespace_preserving_prefix) + r'\s+', line):
                line = line[len(whitespace_preserving_prefix):]
            lines.append(line)
        return '\n'.join(lines)

    def settext(self, section, option, value, whitespace_preserving_prefix='.'):
        lines = []
        for line in value.splitlines():
            if re.match(r'\s+', line):
                line = whitespace_preserving_prefix + line
            lines.append(line)
        self.set(section, option, '\n'.join(lines))

    def write(self, fname, **kw):
        pathlib.Path(fname).write_text(self.write_string(**kw), encoding='utf-8')



"""
Support for standard dataset/app metadata
"""
import collections
import urllib.parse

import attr

__all__ = ['Publisher', 'License', 'Metadata']


@attr.s
class Publisher:
    name = attr.ib(
        metadata=dict(ldkey="http://xmlns.com/foaf/0.1/name"),
        default=None)
    place = attr.ib(
        metadata=dict(ldkey="dc:Location"),
        default=None)
    url = attr.ib(
        metadata=dict(ldkey="http://xmlns.com/foaf/0.1/homepage"),
        default=None)
    contact = attr.ib(
        metadata=dict(ldkey="http://xmlns.com/foaf/0.1/mbox"),
        default=None)


@attr.s
class License:
    name = attr.ib(
        default="Creative Commons Attribution 4.0 International License")
    url = attr.ib(
        default="https://creativecommons.org/licenses/by/4.0/")
    icon = attr.ib(
        default="cc-by.png")


@attr.s
class Metadata:
    publisher = attr.ib(default=Publisher(), validator=attr.validators.instance_of(Publisher))
    license = attr.ib(default=License(), validator=attr.validators.instance_of(License))
    url = attr.ib(default=None)
    title = attr.ib(default=None)
    description = attr.ib(default=None)

    @classmethod
    def from_jsonld(cls, d, defaults=None):
        defaults = defaults or {}
        kw = {}
        for k, v in [
            ('dcat:accessURL', 'url'),
            ('dc:title', 'title'),
            ('dc:description', 'description'),
        ]:
            val = d.get(k) or defaults.get(v)
            if val:
                kw[v] = val
        for ldkey, cls_ in [('dc:publisher', Publisher), ('dc:license', License)]:
            ckw = {}
            dd = d.get(ldkey, {})
            for f in attr.fields(cls_):
                ckw[f.name] = dd.get(f.metadata.get('ldkey', f.name)) \
                    or defaults.get('{0}.{1}'.format(ldkey.split(':')[1], f.name))
            kw[cls_.__name__.lower()] = cls_(**{k: v for k, v in ckw.items() if v})
        return cls(**kw)

    def to_jsonld(self):
        items = [("@context", ["http://www.w3.org/ns/csvw", {"@language": "en"}])]
        for k, v in [
            ('dcat:accessURL', 'url'),
            ('dc:title', 'title'),
            ('dc:description', 'description'),
        ]:
            if getattr(self, v):
                items.append((k, getattr(self, v)))
        for ldkey, cls_ in [('dc:publisher', Publisher), ('dc:license', License)]:
            obj = getattr(self, ldkey.split(':')[1])
            json = collections.OrderedDict()
            for f in attr.fields(cls_):
                if getattr(obj, f.name):
                    json[f.metadata.get('ldkey', f.name)] = getattr(obj, f.name)
            items.append((ldkey, json))
        return collections.OrderedDict(items)

    @property
    def domain(self):
        return urllib.parse.urlparse(self.url).netloc



import os
import sys
import mmap
import shutil
import hashlib
import pathlib
import tempfile
import importlib
import contextlib
import subprocess
import unicodedata

from clldutils.misc import deprecated

Path = pathlib.Path


@contextlib.contextmanager
def sys_path(p):
    sys.path.insert(0, str(Path(p)))
    yield
    sys.path.pop(0)


@contextlib.contextmanager
def memorymapped(filename, access=mmap.ACCESS_READ):
    f = Path(filename).open('rb')
    try:
        m = mmap.mmap(f.fileno(), 0, access=access)
        try:
            yield m
        finally:
            m.close()
    finally:
        f.close()


def import_module(p):
    with sys_path(p.parent):
        m = importlib.import_module(p.stem)
        if Path(m.__file__).parent not in [p.parent, p]:
            # If we end up importing from the wrong place, raise an error:
            raise ImportError(m.__file__)  # pragma: no cover
        return m


# In python 3, pathlib treats path components and string-like representations or
# attributes of paths (like name and stem) as unicode strings. Unfortunately this is not
# true for pathlib under python 2.7. So as workaround for the case of using non-ASCII
# path names with python 2.7 the following two wrapper functions are provided.
# Note that the issue is even more complex, because pathlib with python 2.7 under windows
# may still pose problems.
def path_component(s, encoding='utf-8'):
    if isinstance(s, bytes):
        s = s.decode(encoding)
    return s


def as_unicode(p, encoding='utf-8'):
    deprecated("Use of deprecated function as_unicode! Use str() instead.")
    return '%s' % p


def as_posix(p):
    if hasattr(p, 'as_posix'):
        return p.as_posix()
    elif isinstance(p, str):
        return Path(p).as_posix()
    raise ValueError(p)


def remove(p):
    deprecated('Use of deprecated function remove! Use Path.unlink instead.')
    Path(p).unlink()


def read_text(p, encoding='utf8', **kw):
    deprecated("Use of deprecated function read_text! Use Path.read_text instead.")
    with Path(p).open(encoding=encoding, **kw) as fp:
        return fp.read()


def write_text(p, text, encoding='utf8', **kw):
    deprecated("Use of deprecated function write_text! Use Path.write_text instead.")
    with Path(p).open('w', encoding=encoding, **kw) as fp:
        return fp.write(text)


def readlines(p,
              encoding=None,
              strip=False,
              comment=None,
              normalize=None,
              linenumbers=False):
    """
    Read a `list` of lines from a text file.

    :param p: File path (or `list` or `tuple` of text)
    :param encoding: Registered codec.
    :param strip: If `True`, strip leading and trailing whitespace.
    :param comment: String used as syntax to mark comment lines. When not `None`, \
    commented lines will be stripped. This implies `strip=True`.
    :param normalize: 'NFC', 'NFKC', 'NFD', 'NFKD'
    :param linenumbers: return also line numbers.
    :return: `list` of text lines or pairs (`int`, text or `None`).
    """
    if comment:
        strip = True
    if isinstance(p, (list, tuple)):
        res = [line.decode(encoding) if encoding else line for line in p]
    else:
        with Path(p).open(encoding=encoding or 'utf-8') as fp:
            res = fp.readlines()
    if strip:
        res = [line.strip() or None for line in res]
    if comment:
        res = [None if line and line.startswith(comment) else line for line in res]
    if normalize:
        res = [unicodedata.normalize(normalize, line) if line else line for line in res]
    if linenumbers:
        return [(n, line) for n, line in enumerate(res, 1)]
    return [line for line in res if line is not None]


def rmtree(p, **kw):
    return shutil.rmtree(str(p), **kw)


def move(src, dst):
    return shutil.move(str(src), str(dst))


def copy(src, dst):
    return shutil.copy(str(src), str(dst))


def copytree(src, dst, **kw):
    return shutil.copytree(str(src), str(dst), **kw)


def walk(p, mode='all', **kw):
    """Wrapper for `os.walk`, yielding `Path` objects.

    :param p: root of the directory tree to walk.
    :param mode: 'all|dirs|files', defaulting to 'all'.
    :param kw: Keyword arguments are passed to `os.walk`.
    :return: Generator for the requested Path objects.
    """
    for dirpath, dirnames, filenames in os.walk(str(p), **kw):
        if mode in ('all', 'dirs'):
            for dirname in dirnames:
                yield Path(dirpath).joinpath(dirname)
        if mode in ('all', 'files'):
            for fname in filenames:
                yield Path(dirpath).joinpath(fname)


def md5(p, bufsize=32768):
    hash_md5 = hashlib.md5()
    with Path(p).open('rb') as fp:
        for chunk in iter(lambda: fp.read(bufsize), b''):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


class Manifest(dict):
    """A `dict` mapping relative path names to md5 sums of file contents.

    A `Manifest.from_dir(d, relative_to=d.parent).__unicode__()` is equivalent
    to the content of the file `manifest-md5.txt` of the BagIt specification.

    .. seealso:: https://en.wikipedia.org/wiki/BagIt
    """

    @classmethod
    def from_dir(cls, d, relative_to=None):
        d = Path(d)
        assert d.is_dir()
        return cls((str(p.relative_to(relative_to or d)), md5(p)) for p in walk(d, mode='files'))

    def __str__(self):
        return '\n'.join('{0}  {1}'.format(v, k) for k, v in sorted(self.items()))

    def write(self, outdir=None):
        Path(outdir or '.').joinpath('manifest-md5.txt').write_text(
            '{0}'.format(self), encoding='utf8')


def git_describe(dir_, git_command='git'):
    dir_ = Path(dir_)
    if not dir_.exists():
        raise ValueError('cannot describe non-existent directory')
    dir_ = dir_.resolve()
    cmd = [git_command, '--git-dir=%s' % dir_.joinpath('.git'), 'describe', '--always', '--tags']
    try:
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()
        if p.returncode == 0:
            res = stdout.strip()  # pragma: no cover
        else:
            raise ValueError(stderr)
    except (ValueError, FileNotFoundError):
        res = dir_.name
    if not isinstance(res, str):
        res = res.decode('utf8')
    return res


class TemporaryDirectory(tempfile.TemporaryDirectory):
    def __enter__(self):
        return Path(self.name)



import logging
import pkgutil
import pathlib
import argparse
import warnings
import importlib
import collections
import pkg_resources

import tabulate

from clldutils.loglib import Logging, get_colorlog
from clldutils import markup

__all__ = [
    'ParserError', 'Command', 'command', 'ArgumentParser', 'ArgumentParserWithLogging',
    'get_parser_and_subparsers', 'register_subcommands', 'add_format', 'Table', 'PathType',
]


class ParserError(Exception):
    pass


# Global registry for commands.
# Note: This registry is global so it can only be used for one ArgumentParser instance.
# Otherwise, different ArgumentParsers will share the same sub-commands which will rarely
# be intended.
_COMMANDS = []


class Command(object):

    def __init__(self, func, name=None, usage=None):
        self.func = func
        self.name = name or func.__name__
        self.usage = usage

    @property
    def doc(self):
        return self.usage or self.func.__doc__

    def __call__(self, args):
        return self.func(args)


def command(name=None, usage=None):
    def wrap(f):
        _COMMANDS.append(Command(f, name=name, usage=usage))
        return f
    return wrap


def _attr(obj, attr):
    return getattr(obj, attr, getattr(obj, '__{0}__'.format(attr), None))


class ArgumentParser(argparse.ArgumentParser):
    """A command line argument parser supporting sub-commands in a simple way.

    Sub-commands can be registered in one of two ways:
    - Passing functions as positional arguments into `ArgumentParser.__init__`.
    - Decorating functions with the `command` decorator.
    """

    def __init__(self, pkg_name, *commands, **kw):
        commands = commands or _COMMANDS
        kw.setdefault(
            'description', "Main command line interface of the %s package." % pkg_name)
        kw.setdefault(
            'epilog', "Use '%(prog)s help <cmd>' to get help about individual commands.")
        super(ArgumentParser, self).__init__(**kw)
        self.commands = collections.OrderedDict((_attr(cmd, 'name'), cmd) for cmd in commands)
        self.pkg_name = pkg_name
        self.add_argument("--verbosity", help="increase output verbosity")
        self.add_argument('command', help=' | '.join(self.commands))
        self.add_argument('args', nargs=argparse.REMAINDER)

    def main(self, args=None, catch_all=False, parsed_args=None):
        args = parsed_args or self.parse_args(args=args)
        if args.command == 'help' and len(args.args):
            # As help text for individual commands we simply re-use the docstrings of the
            # callables registered for the command:
            print(_attr(self.commands[args.args[0]], 'doc'))
        else:
            if args.command not in self.commands:
                print('invalid command')
                self.print_help()
                return 64
            try:
                self.commands[args.command](args)
            except ParserError as e:
                print(e)
                print(_attr(self.commands[args.command], 'doc'))
                return 64
            except Exception as e:
                if catch_all:
                    print(e)
                    return 1
                raise
        return 0


class ArgumentParserWithLogging(ArgumentParser):

    def __init__(self, pkg_name, *commands, **kw):
        super(ArgumentParserWithLogging, self).__init__(pkg_name, *commands, **kw)
        self.add_argument('--log', default=get_colorlog(pkg_name), help=argparse.SUPPRESS)
        self.add_argument(
            '--log-level',
            default=logging.INFO,
            help='log level [ERROR|WARN|INFO|DEBUG]',
            type=lambda x: getattr(logging, x))

    def main(self, args=None, catch_all=False, parsed_args=None):
        args = parsed_args or self.parse_args(args=args)
        with Logging(args.log, level=args.log_level):
            return super(ArgumentParserWithLogging, self).main(
                catch_all=catch_all, parsed_args=args)


def confirm(question, default=True):
    """Ask a yes/no question interactively.

    :param question: The text of the question to ask.
    :returns: True if the answer was "yes", False otherwise.
    """
    valid = {"": default, "yes": True, "y": True, "no": False, "n": False}
    while 1:
        choice = input(question + (" [Y/n] " if default else " [y/N] ")).lower()
        if choice in valid:
            return valid[choice]
        print("Please respond with 'y' or 'n' ")


class Formatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    pass


def get_parser_and_subparsers(prog, with_defaults_help=True, with_log=True):
    kw = dict(prog=prog)
    if with_defaults_help:
        kw.update(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser = argparse.ArgumentParser(**kw)
    if with_log:
        parser.add_argument(
            '--log',
            default=get_colorlog(prog),
            help=argparse.SUPPRESS)
        parser.add_argument(
            '--log-level',
            default=logging.INFO,
            help='log level [ERROR|WARN|INFO|DEBUG]',
            type=lambda x: getattr(logging, x))

    subparsers = parser.add_subparsers(
        title="available commands",
        dest="_command",
        description='Run "COMAMND -h" to get help for a specific command.',
        metavar="COMMAND")
    return parser, subparsers


def iter_modules(pkg):
    """ Autodiscover and import all modules in a package.
    """
    for _, name, ispkg in pkgutil.iter_modules(pkg.__path__):
        if not ispkg:
            modname = ".".join([pkg.__name__, name])
            try:
                yield name, importlib.import_module(modname)
            except Exception as e:  # pragma: no cover
                warnings.warn('{0} {1}'.format(e, modname))


def register_subcommands(subparsers, pkg, entry_point=None, formatter_class=Formatter):
    # Discover available commands:
    # Commands are identified by (<entry point name>).<module name>
    _cmds = collections.OrderedDict()
    _cmds.update(list(iter_modules(pkg)))
    if entry_point:
        # ... then look for commands provided in other packages:
        for ep in pkg_resources.iter_entry_points(entry_point):
            try:
                pkg = ep.load()
            except ImportError:
                warnings.warn('ImportError loading entry point {0.name}'.format(ep))
                continue
            _cmds.update(
                [('.'.join([ep.name, name]), mod) for name, mod in iter_modules(pkg)])

    for name, mod in _cmds.items():
        if not mod.__doc__:
            raise ValueError('Command \"{0}\" is missing a docstring.'.format(name))

        subparser = subparsers.add_parser(
            name,
            help=mod.__doc__.strip().splitlines()[0] if mod.__doc__.strip() else '',
            description=mod.__doc__,
            formatter_class=formatter_class)
        if hasattr(mod, 'register'):
            mod.register(subparser)
        subparser.set_defaults(main=mod.run)

    return _cmds


def add_format(parser, default='pipe'):
    parser.add_argument(
        "--format",
        default=default,
        choices=tabulate.tabulate_formats,
        help="Format of tabular output.")


class Table(markup.Table):
    def __init__(self, args, *cols, **kw):
        kw.setdefault('tablefmt', args.format)
        super().__init__(*cols, **kw)


class PathType(object):
    """
    A type to parse `pathlib.Path` instances from the command line.

    Similar to `argparse.FileType`.
    """
    def __init__(self, must_exist=True, type=None):
        assert type in (None, 'dir', 'file')
        self._must_exist = must_exist
        self._type = type

    def __call__(self, string):
        p = pathlib.Path(string)
        if self._must_exist and not p.exists():
            raise argparse.ArgumentTypeError('Path {0} does not exist!'.format(string))
        if p.exists() and self._type and not getattr(p, 'is_' + self._type)():
            raise argparse.ArgumentTypeError('Path {0} is not a {1}!'.format(string, self._type))
        return p



import logging

import colorlog


def get_colorlog(name, stream=None, level=logging.INFO):
    logging.basicConfig(level=level)
    handler = colorlog.StreamHandler(stream)
    handler.setFormatter(colorlog.ColoredFormatter(
        '%(log_color)s%(levelname)-7s%(reset)s %(message)s'))
    log = logging.getLogger(name)
    log.setLevel(level)
    log.propagate = False
    log.addHandler(handler)
    return log


class Logging(object):
    """A context manager to execute a block of code at a specific logging level."""

    def __init__(self, logger, level=logging.DEBUG):
        self.level = level
        self.logger = logger
        self.prev_level = self.logger.getEffectiveLevel()
        root = logging.getLogger()
        self.root_logger_level = root.getEffectiveLevel()
        self.root_handler_level = root.handlers[0].level

    def __enter__(self):
        self.logger.setLevel(self.level)
        if self.logger.handlers:
            self.logger.handlers[0].setLevel(self.level)
        root = logging.getLogger()
        root.setLevel(self.level)
        if root.handlers:
            root.handlers[0].setLevel(self.level)
        return self.logger

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.setLevel(self.prev_level)
        root = logging.getLogger()
        root.setLevel(self.root_logger_level)
        if root.handlers:
            root.handlers[0].setLevel(self.root_handler_level)



import re
import functools
import collections

import attr

from clldutils.text import PATTERN_TYPE

__all__ = ['asdict', 'valid_range', 'valid_re', 'valid_enum_member', 'cmp_off']

# Avoid deprecation warnings for "cmp=False"
# See https://www.attrs.org/en/stable/api.html#deprecated-apis
if getattr(attr, "__version_info__", (0,)) >= (19, 2):
    cmp_off = {"eq": False}
else:  # pragma: no cover
    cmp_off = {"cmp": False}


def defaults(cls):
    res = collections.OrderedDict()
    for field in attr.fields(cls):
        default = field.default
        if isinstance(default, attr.Factory):
            default = default.factory()
        res[field.name] = default
    return res


def asdict(obj, omit_defaults=True, omit_private=True):
    defs = defaults(obj.__class__)
    res = collections.OrderedDict()
    for field in attr.fields(obj.__class__):
        if not (omit_private and field.name.startswith('_')):
            value = getattr(obj, field.name)
            if not (omit_defaults and value == defs[field.name]):
                if hasattr(value, 'asdict'):
                    value = value.asdict(omit_defaults=True)
                res[field.name] = value
    return res


def _valid_enum_member(choices, instance, attribute, value, nullable=False):
    if not (nullable and value is None) and value not in choices:
        raise ValueError('{0} is not a valid {1}'.format(value, attribute.name))


def valid_enum_member(choices, nullable=False):
    return functools.partial(_valid_enum_member, choices, nullable=nullable)


def _valid_range(min_, max_, instance, attribute, value, nullable=False):
    if not (nullable and value is None) and (
            (min_ is not None and value < min_) or (max_ is not None and value > max_)):
        raise ValueError('{0} is not a valid {1}'.format(value, attribute.name))


def valid_range(min_, max_, nullable=False):
    return functools.partial(_valid_range, min_, max_, nullable=nullable)


def _valid_re(regex, instance, attribute, value, nullable=False):
    if nullable and value is None:
        return
    if not isinstance(regex, PATTERN_TYPE):
        regex = re.compile(regex)
    if not regex.match(value):
        raise ValueError('{0} is not a valid {1}'.format(value, attribute.name))


def valid_re(regex, nullable=False):
    return functools.partial(_valid_re, regex, nullable=nullable)



"""
Functionality to create/drop and use databases specified by DB URL.
"""
import shutil
import pathlib
import sqlite3
import subprocess
import urllib.parse

__all__ = ['DB', 'TempDB', 'FreshDB']

CREATEDB = 'createdb'
DROPDB = 'dropdb'
PSQL = 'psql'


class DB:
    """
    A relational database specified by DB URL. Supported dialects are "sqlite" and "postgresql".

    See also https://docs.sqlalchemy.org/en/13/core/engines.html#sqlalchemy.create_engine
    """
    settings_key = 'sqlalchemy.url'

    def __init__(self, url, log=None):
        self.log = log
        self.components = urllib.parse.urlparse(url)
        if self.dialect not in ['sqlite', 'postgresql']:
            raise NotImplementedError(self.dialect)
        if self.dialect == 'postgresql':
            for cmd in [CREATEDB, DROPDB, PSQL]:
                if not shutil.which(cmd):
                    raise ValueError('dialect postgresql requires command {0}'.format(cmd))

    def __str__(self):
        return self.components.geturl()

    @classmethod
    def from_settings(cls, settings, log=None):
        """
        Instantiate a DB looking up the URL in a `dict`.

        :param settings: A `dict` as returned - e.g. - by `pyramid.paster.get_appsettings`.
        """
        return cls(settings[cls.settings_key], log=log)

    @property
    def dialect(self):
        return self.components.scheme.split('+')[0]

    @property
    def name(self):
        assert self.components.path.startswith('/')
        return self.components.path[1:].split('?')[0]

    def exists(self):
        if self.dialect == 'postgresql':
            dbs = [
                line.split('|')[0].strip() for line in
                subprocess.check_output([PSQL, '-lqt']).decode('utf8').splitlines()]
            return self.name in dbs
        return pathlib.Path(self.name).exists()

    def create(self):
        if self.log:
            self.log.info('creating {0}'.format(self))
        if self.dialect == 'postgresql':
            subprocess.check_call([CREATEDB, self.name])
        else:  # self.dialect == 'sqlite'
            if self.exists():
                raise ValueError('db exists!')
            sqlite3.connect(self.name)

    def drop(self):
        if self.exists():
            if self.log:
                self.log.info('dropping {0}'.format(self))
            if self.dialect == 'postgresql':
                subprocess.check_call([DROPDB, self.name])
            else:
                pathlib.Path(self.name).unlink()


class FreshDB(DB):
    """
    Use a newly created database.

    >>> with FreshDB(url) as db:
    ...     assert db.exists()
    """
    def __enter__(self):
        self.drop()
        self.create()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class TempDB(DB):
    """
    Use a temporary database.

    >>> with TempDB(url) as db:
    ...     assert db.exists()
    >>> assert not db.exists()
    """
    def __enter__(self):
        assert not self.exists()
        self.create()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.drop()



"""From "The Enum Recipe": http://techspot.zzzeek.org/2011/01/14/the-enum-recipe/"""

import functools


@functools.total_ordering
class EnumSymbol(object):
    """Define a fixed symbol tied to a parent class."""

    def __init__(self, cls_, name, value, description, *args):
        self.cls_ = cls_
        self.name = name
        self.value = value
        self.description = description
        self.args = args

    def __reduce__(self):
        """Allow unpickling to return the symbol linked to the DeclEnum class."""
        return getattr, (self.cls_, self.name)  # pragma: no cover

    def __iter__(self):
        return iter([self.value, self.description])

    def __repr__(self):
        return "<%s>" % self.name

    def __hash__(self):
        return self.value

    def __str__(self):
        return '{0}'.format(self.value)

    def __lt__(self, other):
        return self.value < getattr(other, 'value', None)

    def __json__(self, *args, **kw):
        return self.value


class EnumMeta(type):
    """Generate new DeclEnum classes."""

    def __init__(cls, classname, bases, dict_):
        cls._reg = reg = cls._reg.copy()
        for k, v in dict_.items():
            if isinstance(v, tuple):
                sym = reg[v[0]] = EnumSymbol(cls, k, *v)
                setattr(cls, k, sym)
        super(EnumMeta, cls).__init__(classname, bases, dict_)

    def __iter__(cls):
        return iter(sorted(cls._reg.values()))


class DeclEnum(metaclass=EnumMeta):
    """Declarative enumeration."""

    _reg = {}

    @classmethod
    def from_string(cls, value):
        try:
            return cls._reg[value]
        except KeyError:
            raise ValueError("Invalid value for %r: %r" % (cls.__name__, value))

    @classmethod
    def get(cls, item):
        if item in iter(cls):
            return item
        for li in cls:
            if li.name == item or li.value == item:
                return li
        raise ValueError(item)

    @classmethod
    def values(cls):
        return list(cls._reg)



import re
import textwrap

from clldutils.misc import nfilter, deprecated

# Brackets are pairs of single characters (<start-token>, <end-token>):
BRACKETS = {
    "(": ")",
    "{": "}",
    "[": "]",
    "（": "）",
    "【": "】",
    "『": "』",
    "«": "»",
    "⁽": "⁾",
    "₍": "₎"
}
# To make it possible to detect compiled regex patterns, we store their type.
# See also http://stackoverflow.com/a/6102100
PATTERN_TYPE = type(re.compile('a'))
# A string of all unicode characters regarded as whitespace (by python's re module \s):
# See also http://stackoverflow.com/a/37903645
WHITESPACE = \
    '\t\n\x0b\x0c\r\x1c\x1d\x1e\x1f \x85\xa0\u1680\u2000\u2001\u2002\u2003\u2004\u2005' \
    '\u2006\u2007\u2008\u2009\u200a\u2028\u2029\u202f\u205f\u3000'


class TextType(object):

    text = 1  # token outside of brackets
    open = 2  # start-token of a bracket
    context = 3  # non-bracket token inside brackets
    close = 4  # end-token of a bracket


def _tokens(text, brackets=None):
    if brackets is None:
        brackets = BRACKETS
    stack = []
    for c in text:
        if c in brackets:
            stack.append(brackets[c])
            yield c, TextType.open
        elif stack and c == stack[-1]:
            stack.pop()
            yield c, TextType.close
        elif not stack:
            yield c, TextType.text
        else:
            yield c, TextType.context


def strip_brackets(text, brackets=None):
    """Strip brackets and what is inside brackets from text.

    .. note::
        If the text contains only one opening bracket, the rest of the text
        will be ignored. This is a feature, not a bug, as we want to avoid that
        this function raises errors too easily.
    """
    res = []
    for c, type_ in _tokens(text, brackets=brackets):
        if type_ == TextType.text:
            res.append(c)
    return ''.join(res).strip()


def split_text_with_context(text, separators=WHITESPACE, brackets=None):
    """Splits text at separators outside of brackets.

    :param text:
    :param separators: An iterable of single character tokens.
    :param brackets:
    :return: A `list` of non-empty chunks.

    .. note:: This function leaves content in brackets in the chunks.
    """
    res, chunk = [], []
    for c, type_ in _tokens(text, brackets=brackets):
        if type_ == TextType.text and c in separators:
            res.append(''.join(chunk).strip())
            chunk = []
        else:
            chunk.append(c)
    res.append(''.join(chunk).strip())
    return nfilter(res)


def split_text(text, separators=re.compile(r'\s'), brackets=None, strip=False):
    """Split text along the separators unless they appear within brackets.

    :param separators: An iterable single characters or a compiled regex pattern.
    :param brackets: `dict` mapping start tokens to end tokens of what is to be \
    recognized as brackets.

    .. note:: This function will also strip content within brackets.
    """
    if not isinstance(separators, PATTERN_TYPE):
        separators = re.compile(
            r'[{0}]'.format(''.join(r'\{0}'.format(c) for c in separators)))

    return nfilter(
        s.strip() if strip else s for s in
        separators.split(strip_brackets(text, brackets=brackets)))


def strip_chars(chars, sequence):
    """Strip the specified chars from anywhere in the text.

    :param chars: An iterable of single character tokens to be stripped out.
    :param sequence: An iterable of single character tokens.
    :return: Text string concatenating all tokens in sequence which were not stripped.
    """
    return ''.join(s for s in sequence if s not in chars)


def truncate_with_ellipsis(t, ellipsis='\u2026', width=40, **kw):
    deprecated('Use of deprecated function truncate_with_ellipsis! Use textwrap.shorten instead.')
    return textwrap.shorten(t, placeholder=ellipsis, width=width, **kw)



import re
import json
import pathlib
import datetime
import contextlib
import collections

import dateutil.parser

DATETIME_ISO_FORMAT = re.compile(r'[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]+')


def parse(d):
    """Convert iso formatted timestamps found as values in the dict d to datetime objects.

    :return: A shallow copy of d with converted timestamps.
    """
    res = {}
    for k, v in d.items():
        if isinstance(v, str) and DATETIME_ISO_FORMAT.match(v):
            v = dateutil.parser.parse(v)
        res[k] = v
    return res


def format(value):
    if isinstance(value, (datetime.date, datetime.datetime)):
        return value.isoformat()
    return value


def dump(obj, path, **kw):
    """json.dump which understands filenames.

    :param obj: The object to be dumped.
    :param path: The path of the JSON file to be written.
    :param kw: Keyword parameters are passed to json.dump
    """
    # avoid indented lines ending with ", " on PY2
    if kw.get('indent') and kw.get('separators') is None:
        kw['separators'] = (',', ': ')

    with pathlib.Path(path).open('w', encoding='utf-8') as fp:
        return json.dump(obj, fp, **kw)


def load(path, **kw):
    """json.load which understands filenames.

    :param kw: Keyword parameters are passed to json.load
    :return: The python object read from path.
    """
    with pathlib.Path(path).open(encoding='utf-8') as fp:
        return json.load(fp, **kw)


@contextlib.contextmanager
def update(path, default=None, load_kw=None, **kw):
    path = pathlib.Path(path)
    if not path.exists():
        if default is None:
            raise ValueError('path does not exist')
        res = default
    else:
        res = load(path, **(load_kw or {}))
    yield res
    dump(res, path, **kw)


def update_ordered(path, **kw):
    return update(
        path,
        default=collections.OrderedDict(),
        load_kw=dict(object_pairs_hook=collections.OrderedDict),
        **kw)



"""Functionality to handle SIL Standard Format (SFM) files

#
# FIXME: seealso link for SFM spec!
#

This format is used natively for Toolbox. Applications which can export in a SFM format
include
- ELAN   FIXME: link!
- Flex   FIXME: link!

This implementation supports
- multiline values
- custom entry separator
"""

import re
import pathlib
import collections

MARKER_PATTERN = re.compile(r'\\(?P<marker>[A-Za-z0-9][A-Za-z0-9_]*)(\s+|$)')

FIELD_SPLITTER_PATTERN = re.compile(r';\s+')


def marker_split(block):
    """Yield marker, value pairs from a text block (i.e. a list of lines).

    :param block: text block consisting of \n separated lines as it will be the case for \
    files read using "rU" mode.
    :return: generator of (marker, value) pairs.
    """
    marker = None
    value = []

    for line in block.split('\n'):
        line = line.strip()
        if line.startswith('\\_'):
            continue  # we simply ignore SFM header fields
        match = MARKER_PATTERN.match(line)
        if match:
            if marker:
                yield marker, '\n'.join(value)
            marker = match.group('marker')
            value = [line[match.end():]]
        else:
            value.append(line)
    if marker:
        yield marker, ('\n'.join(value)).strip()


class Entry(list):
    """We store entries in SFM files as lists of (marker, value) pairs."""

    @classmethod
    def from_string(cls, block, keep_empty=False):
        entry = cls()
        for marker, value in marker_split(block.strip()):
            value = value.strip()
            if value or keep_empty:
                entry.append((marker, value))
        return entry

    def markers(self):
        return collections.Counter(k for k, _ in self)

    def get(self, key, default=None):
        """Retrieve the first value for a marker or None."""
        for k, v in self:
            if k == key:
                return v
        return default

    def getall(self, key):
        """Retrieve all values for a marker."""
        return [v for k, v in self if k == key]

    def __str__(self):
        lines = []
        for key, value in self:
            lines.append('%s %s' % (key, value))
        return '\n'.join('\\' + line for line in lines)


def parse(filename, encoding, entry_sep, entry_prefix, keep_empty=False):
    with pathlib.Path(filename).open('r', encoding=encoding, newline=None) as fp:
        content = fp.read()

    for block in content.split(entry_sep):
        if block.strip():
            block = entry_prefix + block
        else:
            continue  # pragma: no cover
        yield [(k, v.strip())
               for k, v in marker_split(block.strip()) if v.strip() or keep_empty]


class SFM(list):
    """A list of Entries

    Simple usage to normalize a sfm file:

    >>> sfm = SFM.from_file(fname, marker_map={'lexeme': 'lx'})
    >>> sfm.write(fname)
    """

    @classmethod
    def from_file(cls, filename, **kw):
        sfm = cls()
        sfm.read(filename, **kw)
        return sfm

    def read(self,
             filename,
             encoding='utf-8',
             marker_map=None,
             entry_impl=Entry,
             entry_sep='\n\n',
             entry_prefix=None,
             keep_empty=False):
        """Extend the list by parsing new entries from a file.

        :param filename:
        :param encoding:
        :param marker_map: A dict used to map marker names.
        :param entry_impl: Subclass of Entry or None
        :param entry_sep:
        :param entry_prefix:
        """
        marker_map = marker_map or {}
        for entry in parse(
                filename,
                encoding,
                entry_sep,
                entry_prefix or entry_sep,
                keep_empty=keep_empty):
            if entry:
                self.append(entry_impl([(marker_map.get(k, k), v) for k, v in entry]))

    def visit(self, visitor):
        for i, entry in enumerate(self):
            self[i] = visitor(entry) or entry

    def write(self, filename, encoding='utf-8'):
        """Write the list of entries to a file.

        :param filename:
        :param encoding:
        :return:
        """
        with pathlib.Path(filename).open('w', encoding=encoding) as fp:
            for entry in self:
                fp.write(str(entry))
                fp.write('\n\n')



"""Badges for inclusion in markdown docs, etc.

.. seealso:: http://shields.io/
"""
from urllib.parse import urlencode, quote


class Colors(object):

    brightgreen = 'brightgreen'
    green = 'green'
    yellowgreen = 'yellowgreen'
    yellow = 'yellow'
    orange = 'orange'
    red = 'red'
    lightgrey = 'lightgrey'
    blue = 'blue'


def badge(subject, status, color, fmt='svg', markdown=True, label=None, **kw):
    label = label or ': '.join([subject, status])
    url = 'https://img.shields.io/badge/{0}-{1}-{2}.{3}{4}'.format(
        quote(subject), quote(status), color, fmt, '?' + urlencode(kw) if kw else '')
    return '![{0}]({1} "{0}")'.format(label, url) if markdown else url



from setuptools import setup, find_packages

setup(
    name='clldutils',
    version='3.8.0',
    description='Utilities for clld apps',
    long_description=open("README.md").read(),
    long_description_content_type='text/markdown',
    author='Robert Forkel',
    author_email='forkel@shh.mpg.de',
    url='https://github.com/clld/clldutils',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    include_package_data=True,
    install_requires=[
        'python-dateutil',
        'tabulate>=0.7.7',
        'colorlog',
        'attrs>=18.1.0',
        'csvw>=1.0',
    ],
    extras_require={
        'dev': ['flake8', 'wheel', 'twine'],
        'test': [
            'pytest>=5.4',
            'pytest-mock',
            'pytest-cov',
            'coverage>=4.2',
        ],
    },
    license="Apache 2.0",
    zip_safe=False,
    keywords='',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



"""
Leipzig Glossing Rules
======================
Conventions for interlinear morpheme-by-morpheme glosses

About the rules
---------------

The Leipzig Glossing Rules have been developed jointly by the Department of
Linguistics of the Max Planck Institute for Evolutionary Anthropology
(Bernard Comrie, Martin Haspelmath) and by the Department of Linguistics
of the University of Leipzig (Balthasar Bickel). They consist of ten rules for the
"syntax" and "semantics" of interlinear glosses, and an appendix with a
proposed "lexicon" of abbreviated category labels. The rules cover a large part
of linguists' needs in glossing texts, but most authors will feel the need to add
(or modify) certain conventions (especially category labels). Still, it will be
useful to have a standard set of conventions that linguists can refer to, and the
Leipzig Rules are proposed as such to the community of linguists. The Rules
are intended to reflect common usage, and only very few (mostly optional)
innovations are proposed.

Leipzig, last change: May 31, 2015
Further updates will be managed by the Committee of Editors of Linguistics Journals.

Important references:
- Lehmann, Christian. 1982. "Directions for interlinear morphemic translations".
  Folia Linguistica 16: 199-224.
- Croft, William. 2003. Typology and universals. 2nd ed. Cambridge: Cambridge
  University Press, pp. xix-xxv.

The rules
---------
(revised version of February 2008)

Preamble
~~~~~~~~

Interlinear morpheme-by-morpheme glosses give information about the
meanings and grammatical properties of individual words and parts of
words. Linguists by and large conform to certain notational conventions in
glossing, and the main purpose of this document is to make the most widely
used conventions explicit.

Depending on the author's purposes and the readers' assumed background
knowledge, different degrees of detail will be chosen. The current rules
therefore allow some flexibility in various respects, and sometimes alternative
options are mentioned.

The main purpose that is assumed here is the presentation of an example in a
research paper or book. When an entire corpus is tagged, somewhat different
considerations may apply (e.g. one may want to add information about larger
units such as words or phrases; the rules here only allow for information
about morphemes).

It should also be noted that there are often multiple ways of analyzing the
morphological patterns of a language. The glossing conventions do not help
linguists in deciding between them, but merely provide standard ways of
abbreviating possible descriptions. Moreover, glossing is rarely a complete
morphological description, and it should be kept in mind that its purpose is
not to state an analysis, but to give some further possibly relevant information
on the structure of a text or an example, beyond the idiomatic translation.

A remark on the treatment of glosses in data cited from other sources: Glosses
are part of the analysis, not part of the data. When citing an example from a
published source, the gloss may be changed by the author if they prefer
different terminology, a different style or a different analysis.


Rule 1: Word-by-word alignment

Interlinear glosses are left-aligned vertically, word by word, with the example. E.g.

(1) Indonesian (Sneddon 1996:237)
    Mereka  di  Jakarta sekarang.
    They    in  Jakarta now
    'They are in Jakarta now.'


Rule 2: Morpheme-by-morpheme correspondence

Segmentable morphemes are separated by hyphens, both in the example and
in the gloss. There must be exactly the same number of hyphens in the
example and in the gloss. E.g.

(2) Lezgian (Haspelmath 1993:207)
    Gila    abur-u-n        ferma   hamišaluǧ   güǧüna  amuq’-da-č.
    now     they-OBL-GEN    farm    forever     behind  stay-FUT-NEG
    ‘Now their farm will not stay behind forever.’

Since hyphens and vertical alignment make the text look unusual, authors
may want to add another line at the beginning, containing the unmodified
text, or resort to the option described in Rule 4 (and especially 4C).
Clitic boundaries are marked by an equals sign, both in the object
language and in the gloss.

(3) West Greenlandic (Fortescue 1984:127)
    palasi=lu   niuirtur=lu
    priest=and  shopkeeper=and
    'both the priest and the shopkeeper'

Rule 2A. (Optional)

If morphologically bound elements constitute distinct prosodic or
phonological words, a hyphen and a single space may be used together in the
object language (but not in the gloss).

(4) Hakha Lai
    a-nii -láay
    3SG-laugh-FUT
    's/he will laugh'


Rule 3: Grammatical category labels

Grammatical morphemes are generally rendered by abbreviated grammatical
category labels, printed in upper case letters (usually small capitals). A list of
standard abbreviations (which are widely known among linguists) is given at
the end of this document.

Deviations from these standard abbreviations may of course be necessary
in particular cases, e.g. if a category is highly frequent in a language, so that a
shorter abbreviation is more convenient, e.g. CPL (instead of COMPL) for
"completive", PF (instead of PRF) for "perfect", etc. If a category is very rare, it
may be simplest not to abbreviate its label at all.

In many cases, either a category label or a word from the metalanguage is
acceptable. Thus, both of

(5) Russian
    My  s       Marko   poexa-l-i   avtobus-om  v   Peredelkino.
    1PL COM     Marko   go-PST-PL   bus-INS     All Peredelkino.
    we  with    Marko   go-PST-PL   bus-by      to  Peredelkino.
    'Marko and I went to Perdelkino by bus.'


Rule 4: One-to-many correspondences

When a single object-language element is rendered by several metalanguage
elements (words or abbreviations), these are separated by periods. E.g.

(6) Turkish
    çık-mak
    come.out-INF
    'to come out'

(7) Latin
    insul-arum
    island-GEN.PL
    'of the islands'

(8) French
    aux         chevaux
    to.ART.PL   horse.PL
    'to the horses'

(9) German
    unser-n     Väter-n
    our-DAT.PL  father.PL-DAT.PL
    'to our fathers'

(10) Hittite (Lehmann 1982:211)
    n=an        apedani     mehuni      essandu.
    CONN=him    that.DAT.SG time.DAT.SG eat.they.shall
    'They shall celebrate him on that date.' (CONN = connective)

(11) Jaminjung (Schultze-Berndt 2000:92)
    nanggayan   guny-bi-yarluga?
    who         2DU.A.3SG.P-FUT-poke
    'Who do you two want to spear?'

The ordering of the two metalanguage elements may be determined by
various principles that are not easy to generalize over, so no rule will be
provided for this.

There are various reasons for a one-to-many correspondence between
object-language elements and gloss elements. These are conflated by the
uniform use of the period. If one wants to distinguish between them, one may
follow Rules 4A-E.


Rule 4A. (Optional)

If an object-language element is neither formally nor semantically segmentable and only
the metalanguage happens to lack a single-word equivalent, the underscore may be used
instead of the period.

(12) Turkish (cf. 6)
    çık-mak
    come_out-INF
    'to come out'


Rule 4B. (Optional)

If an object-language element is formally unsegmentable but has two clearly
distinguishable meanings or grammatical properties, the semi-colon may be used. E.g.

(13) Latin (cf. 7)
    insul-arum
    island-GEN;PL
    'of the islands'

(14) French
    aux         chevaux
    to;ART;PL   horse;PL
    'to the horses'


Rule 4C. (Optional)

If an object-language element is formally and semantically segmentable, but the author
does not want to show the formal segmentation (because it is irrelevant and/or to keep
the text intact), the colon may be used. E.g.

(15) Hittite (Lehmann 1982:211) (cf. 10)
    n=an        apedani     mehuni      essandu.
    CONN=him    that:DAT;SG time:DAT;SG eat:they:shall
    'They shall celebrate him on that date.'


Rule 4D. (Optional)

If a grammatical property in the object-language is signaled by a
morphophonological change (ablaut, mutation, tone alternation, etc.), the
backslash is used to separate the category label and the rest of the gloss.

(16) German (cf. 9)
    unser-n     Väter-n
    our-DAT.PL  father.PL-DAT.PL
    'to our fathers' (cf. singular Vater)

(17) Irish
    bhris-is
    PST\break-2SG
    'you broke' (cf. nonpast bris-)

(18) Kinyarwanda
    mú-kòrà
    SBJV\1PL-work
    'that we work' (cf. indicative mù-kòrà)


Rule 4E. (Optional)

If a language has person-number affixes that express the agent-like and the
patient-like argument of a transitive verb simultaneously, the symbol ">" may
be used in the gloss to indicate that the first is the agent-like argument and the
second is the patient-like argument.

(19) Jaminjung (Schultze-Berndt 2000:92) (cf. 11)
    nanggayan   guny-bi-yarluga?
    who         2DU>3SG-FUT-poke
    'Who do you two want to spear?'


Rule 5: Person and number labels

Person and number are not separated by a period when they occur in this order. E.g.

(20) Italian
    and-iamo
    go-PRS.1PL (not: go-PRS.1.PL)
    'we go'


Rule 5A. (Optional)

Number and gender markers are very frequent in some languages, especially
when combined with person. Several authors therefore use non-capitalized
shortened abbreviations without a period. If this option is adopted, then the
second gloss is used in (21).

(21) Belhare
    ne-e    a-khim-chi      n-yuNNa
    DEM-LOC that.DAT.SG     3NSG-be.NPST
    DEM-LOC 1sPOSS-house-PL 3ns-be.NPST
    'Here are my houses.''


Rule 6: Non-overt elements

If the morpheme-by-morpheme gloss contains an element that does not
correspond to an overt element in the example, it can be enclosed in square
brackets. An obvious alternative is to include an overt "Ø" in the objectlanguage
text, which is separated by a hyphen like an overt element.

(22) Latin
    puer        or: puer-Ø
    boy[NOM.SG]     boy-NOM.SG
    ‘boy’           ‘boy’


Rule 7: Inherent categories

Inherent, non-overt categories such as gender may be indicated in the gloss, but a
special boundary symbol, the round parenthesis, is used. E.g.

(23) Hunzib (van den Berg 1995:46)
    oz#-di-g    xõxe        m-uq'e-r
    boy-OBL-AD  tree(G4)    G4-bend-PRET
    'Because of the boy the tree bent.' (G4 = 4th gender, AD = adessive, PRET = preterite)


Rule 8: Bipartite elements

Grammatical or lexical elements that consist of two parts which are treated as
distinct morphological entities (e.g. bipartite stems such as Lakhota na-xʔu̧ 'hear') may
be treated in two different ways:

(i) The gloss may simply be repeated:

(24) Lakhota
    na-wíčha-wa-xʔu̧
    hear-3PL.UND-1SG.ACT-hear
    'I hear them' (UND = undergoer, ACT = actor)

(i) The gloss may simply be repeated:

(25) Lakhota
    na-wíčha-wa-xʔu̧
    hear-3PL.UND-1SG.ACT- STEM
    'I hear them'

Circumfixes are "bipartite affixes" and can be treated in the same way, e.g.

(26) German
    ge-seh-en       or: ge-seh-en
    PTCP-see-PTCP       PTCP-see-CIRC
    'seen'              'seen'


Rule 9: Infixes

Infixes are enclosed by angle brackets, and so is the object-language
counterpart in the gloss.

(27) Tagalog
    b<um>ili (stem: bili)
    <ACTFOC>buy
    'buy'

(28) Latin
    reli<n>qu-ere (stem: reliqu-)
    leave<PRS>-INF
    'to leave'

Infixes are generally easily identifiable as left-peripheral (as in 27) or as
rightperipheral (as in 28), and this determines the position of the gloss
corresponding to the infix with respect to the gloss of the stem. If the infix is
not clearly peripheral, some other basis for linearizing the gloss has to be
found.


Rule 10: Reduplication

Reduplication is treated similarly to affixation, but with a tilde (instead of an
ordinary hyphen) connecting the copied element to the stem.

(29) Hebrew
    yerak~rak-im
    green~ATT-M.PL
    'greenish ones' (ATT= attenuative)

(30) Tagalog
    bi~bili
    IPFV~buy
    'is buying'

(31) Tagalog
    b<um>i~bili
    <ACTFOC>IPFV~buy
    'is buying' (ACTFOC = Actor focus)


Appendix: List of Standard Abbreviations

See below.


References
~~~~~~~~~~

- Fortescue, Michael. 1984. West Greenlandic. (Croom Helm descriptive grammars)
  London: Croom Helm.
- Haspelmath, Martin. 1993. A grammar of Lezgian. (Mouton Grammar Library, 9).
  Berlin - New York: Mouton de Gruyter.
- Lehmann, Christian. 1983. "Directions for interlinear morphemic translations".
  Folia Linguistica 16: 193-224.
- Schultze-Berndt, Eva. 2000. Simple and complex verbs in Jaminjung:
  A study of event categorization in an Australian language.
  Katholieke Universiteit Nijmegen Ph.D. Dissertation.
- Sneddon, James Neil. 1996. Indonesian: A comprehensive grammar. London: Routledge.
- van den Berg, Helma. 1995. A Grammar of Hunzib.
  (Lincom Studies in Caucasian Linguistics, 1.) München: Lincom Europa.

.. seealso:: http://www.eva.mpg.de/lingua/resources/glossing-rules.php
"""

import re
import itertools

PERSONS = {
    "1": "first person",
    "2": "second person",
    "3": "third person",
}

ABBRS = {
    "A": "agent-like argument of canonical transitive verb",
    "ABL": "ablative",
    "ABS": "absolutive",
    "ACC": "accusative",
    "ADJ": "adjective",
    "ADV": "adverb(ial)",
    "AGR": "agreement",
    "ALL": "allative",
    "ANTIP": "antipassive",
    "APPL": "applicative",
    "ART": "article",
    "AUX": "auxiliary",
    "BEN": "benefactive",
    "CAUS": "causative",
    "CLF": "classifier",
    "COM": "comitative",
    "COMP": "complementizer",
    "COMPL": "completive",
    "COND": "conditional",
    "COP": "copula",
    "CVB": "converb",
    "DAT": "dative",
    "DECL": "declarative",
    "DEF": "definite",
    "DEM": "demonstrative",
    "DET": "determiner",
    "DIST": "distal",
    "DISTR": "distributive",
    "DU": "dual",
    "DUR": "durative",
    "ERG": "ergative",
    "EXCL": "exclusive",
    "F": "feminine",
    "FOC": "focus",
    "FUT": "future",
    "GEN": "genitive",
    "IMP": "imperative",
    "INCL": "inclusive",
    "IND": "indicative",
    "INDF": "indefinite",
    "INF": "infinitive",
    "INS": "instrumental",
    "INTR": "intransitive",
    "IPFV": "imperfective",
    "IRR": "irrealis",
    "LOC": "locative",
    "M": "masculine",
    "N": "neuter",
    "N-": "non- (e.g. NSG nonsingular, NPST nonpast)",
    "NEG": "negation, negative",
    "NMLZ": "nominalizer/nominalization",
    "NOM": "nominative",
    "OBJ": "object",
    "OBL": "oblique",
    "P": "patient-like argument of canonical transitive verb",
    "PASS": "passive",
    "PFV": "perfective",
    "PL": "plural",
    "POSS": "possessive",
    "PRED": "predicative",
    "PRF": "perfect",
    "PRS": "present",
    "PROG": "progressive",
    "PROH": "prohibitive",
    "PROX": "proximal/proximate",
    "PST": "past",
    "PTCP": "participle",
    "PURP": "purposive",
    "Q": "question particle/marker",
    "QUOT": "quotative",
    "RECP": "reciprocal",
    "REFL": "reflexive",
    "REL": "relative",
    "RES": "resultative",
    "S": "single argument of canonical intransitive verb",
    "SBJ": "subject",
    "SBJV": "subjunctive",
    "SG": "singular",
    "TOP": "topic",
    "TR": "transitive",
    "VOC": "vocative",
}


def pattern(custom=None):
    return re.compile(
        '(?P<pre>^|[^A-Z1-3])(?P<person>{0})?(?P<abbr>{1})(?=$|[^A-Z1-3])'.format(
            '|'.join(re.escape(k) for k in PERSONS),
            '|'.join(re.escape(k) for k in itertools.chain(ABBRS, custom or {}))))


def replace(string, repl=None, custom=None):
    custom = custom or {}
    if repl is None:
        def repl(m):
            res = ''
            if m.group('person'):
                res += PERSONS[m.group('person')] + ' '
            res += custom.get(m.group('abbr'), ABBRS.get(m.group('abbr')))
            return '{0}<{1}>'.format(m.group('pre'), res)

    return pattern(custom).sub(repl, string)



import pathlib

import attr

_LICENSES = {
    "Glide": {
        "name": "3dfx Glide License",
        "url": "http://www.users.on.net/~triforce/glidexp/COPYING.txt",
    },
    "Abstyles": {
        "name": "Abstyles License",
        "url": "https://fedoraproject.org/wiki/Licensing/Abstyles",
    },
    "AFL-1.1": {
        "name": "Academic Free License v1.1",
        "url": "http://opensource.linux-mirror.org/licenses/afl-1.1.txt",
    },
    "AFL-1.2": {
        "name": "Academic Free License v1.2",
        "url": "http://opensource.linux-mirror.org/licenses/afl-1.2.txt",
    },
    "AFL-2.0": {
        "name": "Academic Free License v2.0",
        "url": "http://opensource.linux-mirror.org/licenses/afl-2.0.txt",
    },
    "AFL-2.1": {
        "name": "Academic Free License v2.1",
        "url": "http://opensource.linux-mirror.org/licenses/afl-2.1.txt",
    },
    "AFL-3.0": {
        "name": "Academic Free License v3.0",
        "url": "http://www.opensource.org/licenses/afl-3.0",
    },
    "AMPAS": {
        "name": "Academy of Motion Picture Arts and Sciences BSD",
        "url": "https://fedoraproject.org/wiki/Licensing/BSD#AMPASBSD",
    },
    "APL-1.0": {
        "name": "Adaptive Public License 1.0",
        "url": "http://www.opensource.org/licenses/APL-1.0",
    },
    "Adobe-Glyph": {
        "name": "Adobe Glyph List License",
        "url": "https://fedoraproject.org/wiki/Licensing/MIT#AdobeGlyph",
    },
    "APAFML": {
        "name": "Adobe Postscript AFM License",
        "url": "https://fedoraproject.org/wiki/Licensing/AdobePostscriptAFM",
    },
    "Adobe-2006": {
        "name": "Adobe Systems Incorporated Source Code License Agreement",
        "url": "https://fedoraproject.org/wiki/Licensing/AdobeLicense",
    },
    "AGPL-1.0": {
        "name": "Affero General Public License v1.0",
        "url": "http://www.affero.org/oagpl.html",
    },
    "Afmparse": {
        "name": "Afmparse License",
        "url": "https://fedoraproject.org/wiki/Licensing/Afmparse",
    },
    "Aladdin": {
        "name": "Aladdin Free Public License",
        "url": "http://pages.cs.wisc.edu/~ghost/doc/AFPL/6.01/Public.htm",
    },
    "ADSL": {
        "name": "Amazon Digital Services License",
        "url": "https://fedoraproject.org/wiki/Licensing/AmazonDigitalServicesLicense",
    },
    "AMDPLPA": {
        "name": "AMD's plpa_map.c License",
        "url": "https://fedoraproject.org/wiki/Licensing/AMD_plpa_map_License",
    },
    "ANTLR-PD": {
        "name": "ANTLR Software Rights Notice",
        "url": "http://www.antlr2.org/license.html",
    },
    "Apache-1.0": {
        "name": "Apache License 1.0",
        "url": "http://www.apache.org/licenses/LICENSE-1.0",
    },
    "Apache-1.1": {
        "name": "Apache License 1.1",
        "url": "http://apache.org/licenses/LICENSE-1.1",
    },
    "Apache-2.0": {
        "name": "Apache License 2.0",
        "url": "http://www.apache.org/licenses/LICENSE-2.0",
    },
    "AML": {
        "name": "Apple MIT License",
        "url": "https://fedoraproject.org/wiki/Licensing/Apple_MIT_License",
    },
    "APSL-1.0": {
        "name": "Apple Public Source License 1.0",
        "url": "https://fedoraproject.org/wiki/Licensing/Apple_Public_Source_License_1.0",
    },
    "APSL-1.2": {
        "name": "Apple Public Source License 1.2",
        "url": "http://www.samurajdata.se/opensource/mirror/licenses/apsl.php",
    },
    "APSL-2.0": {
        "name": "Apple Public Source License 2.0",
        "url": "http://www.opensource.apple.com/license/apsl/",
    },
    "Artistic-1.0": {
        "name": "Artistic License 1.0",
        "url": "http://opensource.org/licenses/Artistic-1.0",
    },
    "Artistic-1.0-Perl": {
        "name": "Artistic License 1.0 (Perl)",
        "url": "http://dev.perl.org/licenses/artistic.html",
    },
    "Artistic-1.0-cl8": {
        "name": "Artistic License 1.0 w/clause 8",
        "url": "http://opensource.org/licenses/Artistic-1.0",
    },
    "Artistic-2.0": {
        "name": "Artistic License 2.0",
        "url": "http://www.opensource.org/licenses/artistic-license-2.0",
    },
    "AAL": {
        "name": "Attribution Assurance License",
        "url": "http://www.opensource.org/licenses/attribution",
    },
    "Bahyph": {
        "name": "Bahyph License",
        "url": "https://fedoraproject.org/wiki/Licensing/Bahyph",
    },
    "Barr": {
        "name": "Barr License",
        "url": "https://fedoraproject.org/wiki/Licensing/Barr",
    },
    "Beerware": {
        "name": "Beerware License",
        "url": "https://fedoraproject.org/wiki/Licensing/Beerware",
    },
    "BitTorrent-1.1": {
        "name": "BitTorrent Open Source License v1.1",
        "url": "http://directory.fsf.org/wiki/License:BitTorrentOSL1.1",
    },
    "BSL-1.0": {
        "name": "Boost Software License 1.0",
        "url": "http://www.boost.org/LICENSE_1_0.txt",
    },
    "Borceux": {
        "name": "Borceux license",
        "url": "https://fedoraproject.org/wiki/Licensing/Borceux",
    },
    "BSD-2-Clause": {
        "name": "BSD 2-clause \"Simplified\" License",
        "url": "http://www.opensource.org/licenses/BSD-2-Clause",
    },
    "BSD-2-Clause-FreeBSD": {
        "name": "BSD 2-clause FreeBSD License",
        "url": "http://www.freebsd.org/copyright/freebsd-license.html",
    },
    "BSD-2-Clause-NetBSD": {
        "name": "BSD 2-clause NetBSD License",
        "url": "http://www.netbsd.org/about/redistribution.html#default",
    },
    "BSD-3-Clause": {
        "name": "BSD 3-clause \"New\" or \"Revised\" License",
        "url": "http://www.opensource.org/licenses/BSD-3-Clause",
    },
    "BSD-3-Clause-Clear": {
        "name": "BSD 3-clause Clear License",
        "url": "http://labs.metacarta.com/license-explanation.html#license",
    },
    "BSD-4-Clause": {
        "name": "BSD 4-clause \"Original\" or \"Old\" License",
        "url": "http://directory.fsf.org/wiki/License:BSD_4Clause",
    },
    "BSD-Protection": {
        "name": "BSD Protection License",
        "url": "https://fedoraproject.org/wiki/Licensing/BSD_Protection_License",
    },
    "BSD-3-Clause-Attribution": {
        "name": "BSD with attribution",
        "url": "https://fedoraproject.org/wiki/Licensing/BSD_with_Attribution",
    },
    "0BSD": {
        "name": "BSD Zero Clause License",
        "url": "http://landley.net/toybox/license.html ",
    },
    "BSD-4-Clause-UC": {
        "name": "BSD-4-Clause (University of California-Specific)",
        "url": "http://www.freebsd.org/copyright/license.html",
    },
    "bzip2-1.0.5": {
        "name": "bzip2 and libbzip2 License v1.0.5",
        "url": "http://bzip.org/1.0.5/bzip2-manual-1.0.5.html",
    },
    "bzip2-1.0.6": {
        "name": "bzip2 and libbzip2 License v1.0.6",
        "url": "https://github.com/asimonov-im/bzip2/blob/master/LICENSE",
    },
    "Caldera": {
        "name": "Caldera License",
        "url": "http://www.lemis.com/grog/UNIX/ancient-source-all.pdf",
    },
    "CECILL-1.0": {
        "name": "CeCILL Free Software License Agreement v1.0",
        "url": "http://www.cecill.info/licences/Licence_CeCILL_V1-fr.html",
    },
    "CECILL-1.1": {
        "name": "CeCILL Free Software License Agreement v1.1",
        "url": "http://www.cecill.info/licences/Licence_CeCILL_V1.1-US.html",
    },
    "CECILL-2.0": {
        "name": "CeCILL Free Software License Agreement v2.0",
        "url": "http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html",
    },
    "CECILL-2.1": {
        "name": "CeCILL Free Software License Agreement v2.1",
        "url": "http://opensource.org/licenses/CECILL-2.1",
    },
    "CECILL-B": {
        "name": "CeCILL-B Free Software License Agreement",
        "url": "http://www.cecill.info/licences/Licence_CeCILL-B_V1-fr.html",
    },
    "CECILL-C": {
        "name": "CeCILL-C Free Software License Agreement",
        "url": "http://www.cecill.info/licences/Licence_CeCILL-C_V1-fr.html",
    },
    "ClArtistic": {
        "name": "Clarified Artistic License",
        "url": "http://www.ncftp.com/ncftp/doc/LICENSE.txt",
    },
    "MIT-CMU": {
        "name": "CMU License",
        "url": "https://fedoraproject.org/wiki/Licensing:MIT?rd=Licensing/MIT#CMU_Style",
    },
    "CNRI-Jython": {
        "name": "CNRI Jython License",
        "url": "http://www.jython.org/license.html",
    },
    "CNRI-Python": {
        "name": "CNRI Python License",
        "url": "http://www.opensource.org/licenses/CNRI-Python",
    },
    "CNRI-Python-GPL-Compatible": {
        "name": "CNRI Python Open Source GPL Compatible License Agreement",
        "url": "http://www.python.org/download/releases/1.6.1/download_win/",
    },
    "CPOL-1.02": {
        "name": "Code Project Open License 1.02",
        "url": "http://www.codeproject.com/info/cpol10.aspx",
    },
    "CDDL-1.0": {
        "name": "Common Development and Distribution License 1.0",
        "url": "http://www.opensource.org/licenses/cddl1",
    },
    "CDDL-1.1": {
        "name": "Common Development and Distribution License 1.1",
        "url": "http://glassfish.java.net/public/CDDL+GPL_1_1.html",
    },
    "CPAL-1.0": {
        "name": "Common Public Attribution License 1.0",
        "url": "http://www.opensource.org/licenses/CPAL-1.0",
    },
    "CPL-1.0": {
        "name": "Common Public License 1.0",
        "url": "http://opensource.org/licenses/CPL-1.0",
    },
    "CATOSL-1.1": {
        "name": "Computer Associates Trusted Open Source License 1.1",
        "url": "http://opensource.org/licenses/CATOSL-1.1",
    },
    "Condor-1.1": {
        "name": "Condor Public License v1.1",
        "url": "http://research.cs.wisc.edu/condor/license.html#condor",
    },
    "CC-BY-1.0": {
        "name": "Creative Commons Attribution 1.0",
        "url": "https://creativecommons.org/licenses/by/1.0/",
    },
    "CC-BY-2.0": {
        "name": "Creative Commons Attribution 2.0",
        "url": "https://creativecommons.org/licenses/by/2.0/",
    },
    "CC-BY-2.5": {
        "name": "Creative Commons Attribution 2.5",
        "url": "https://creativecommons.org/licenses/by/2.5/",
    },
    "CC-BY-3.0": {
        "name": "Creative Commons Attribution 3.0",
        "url": "https://creativecommons.org/licenses/by/3.0/",
    },
    "CC-BY-4.0": {
        "name": "Creative Commons Attribution 4.0",
        "url": "https://creativecommons.org/licenses/by/4.0/",
    },
    "CC-BY-ND-1.0": {
        "name": "Creative Commons Attribution No Derivatives 1.0",
        "url": "https://creativecommons.org/licenses/by-nd/1.0/",
    },
    "CC-BY-ND-2.0": {
        "name": "Creative Commons Attribution No Derivatives 2.0",
        "url": "https://creativecommons.org/licenses/by-nd/2.0/",
    },
    "CC-BY-ND-2.5": {
        "name": "Creative Commons Attribution No Derivatives 2.5",
        "url": "https://creativecommons.org/licenses/by-nd/2.5/",
    },
    "CC-BY-ND-3.0": {
        "name": "Creative Commons Attribution No Derivatives 3.0",
        "url": "https://creativecommons.org/licenses/by-nd/3.0/",
    },
    "CC-BY-ND-4.0": {
        "name": "Creative Commons Attribution No Derivatives 4.0",
        "url": "https://creativecommons.org/licenses/by-nd/4.0/",
    },
    "CC-BY-NC-1.0": {
        "name": "Creative Commons Attribution Non Commercial 1.0",
        "url": "https://creativecommons.org/licenses/by-nc/1.0/",
    },
    "CC-BY-NC-2.0": {
        "name": "Creative Commons Attribution Non Commercial 2.0",
        "url": "https://creativecommons.org/licenses/by-nc/2.0/",
    },
    "CC-BY-NC-2.5": {
        "name": "Creative Commons Attribution Non Commercial 2.5",
        "url": "https://creativecommons.org/licenses/by-nc/2.5/",
    },
    "CC-BY-NC-3.0": {
        "name": "Creative Commons Attribution Non Commercial 3.0",
        "url": "https://creativecommons.org/licenses/by-nc/3.0/",
    },
    "CC-BY-NC-4.0": {
        "name": "Creative Commons Attribution Non Commercial 4.0",
        "url": "https://creativecommons.org/licenses/by-nc/4.0/",
    },
    "CC-BY-NC-ND-1.0": {
        "name": "Creative Commons Attribution Non Commercial No Derivatives 1.0",
        "url": "https://creativecommons.org/licenses/by-nd-nc/1.0/",
    },
    "CC-BY-NC-ND-2.0": {
        "name": "Creative Commons Attribution Non Commercial No Derivatives 2.0",
        "url": "https://creativecommons.org/licenses/by-nc-nd/2.0/",
    },
    "CC-BY-NC-ND-2.5": {
        "name": "Creative Commons Attribution Non Commercial No Derivatives 2.5",
        "url": "https://creativecommons.org/licenses/by-nc-nd/2.5/",
    },
    "CC-BY-NC-ND-3.0": {
        "name": "Creative Commons Attribution Non Commercial No Derivatives 3.0",
        "url": "https://creativecommons.org/licenses/by-nc-nd/3.0/",
    },
    "CC-BY-NC-ND-4.0": {
        "name": "Creative Commons Attribution Non Commercial No Derivatives 4.0",
        "url": "https://creativecommons.org/licenses/by-nc-nd/4.0/",
    },
    "CC-BY-NC-SA-1.0": {
        "name": "Creative Commons Attribution Non Commercial Share Alike 1.0",
        "url": "https://creativecommons.org/licenses/by-nc-sa/1.0/",
    },
    "CC-BY-NC-SA-2.0": {
        "name": "Creative Commons Attribution Non Commercial Share Alike 2.0",
        "url": "https://creativecommons.org/licenses/by-nc-sa/2.0/",
    },
    "CC-BY-NC-SA-2.5": {
        "name": "Creative Commons Attribution Non Commercial Share Alike 2.5",
        "url": "https://creativecommons.org/licenses/by-nc-sa/2.5/",
    },
    "CC-BY-NC-SA-3.0": {
        "name": "Creative Commons Attribution Non Commercial Share Alike 3.0",
        "url": "https://creativecommons.org/licenses/by-nc-sa/3.0/",
    },
    "CC-BY-NC-SA-4.0": {
        "name": "Creative Commons Attribution Non Commercial Share Alike 4.0",
        "url": "https://creativecommons.org/licenses/by-nc-sa/4.0/",
    },
    "CC-BY-SA-1.0": {
        "name": "Creative Commons Attribution Share Alike 1.0",
        "url": "https://creativecommons.org/licenses/by-sa/1.0/",
    },
    "CC-BY-SA-2.0": {
        "name": "Creative Commons Attribution Share Alike 2.0",
        "url": "https://creativecommons.org/licenses/by-sa/2.0/",
    },
    "CC-BY-SA-2.5": {
        "name": "Creative Commons Attribution Share Alike 2.5",
        "url": "https://creativecommons.org/licenses/by-sa/2.5/",
    },
    "CC-BY-SA-3.0": {
        "name": "Creative Commons Attribution Share Alike 3.0",
        "url": "https://creativecommons.org/licenses/by-sa/3.0/",
    },
    "CC-BY-SA-4.0": {
        "name": "Creative Commons Attribution Share Alike 4.0",
        "url": "https://creativecommons.org/licenses/by-sa/4.0/",
    },
    "CC0-1.0": {
        "name": "Creative Commons Zero v1.0 Universal",
        "url": "https://creativecommons.org/publicdomain/zero/1.0/",
    },
    "Crossword": {
        "name": "Crossword License",
        "url": "https://fedoraproject.org/wiki/Licensing/Crossword",
    },
    "CUA-OPL-1.0": {
        "name": "CUA Office Public License v1.0",
        "url": "http://opensource.org/licenses/CUA-OPL-1.0",
    },
    "Cube": {
        "name": "Cube License",
        "url": "https://fedoraproject.org/wiki/Licensing/Cube",
    },
    "D-FSL-1.0": {
        "name": "Deutsche Freie Software Lizenz",
        "url": "http://www.dipp.nrw.de/d-fsl/index_html/lizenzen/de/D-FSL-1_0_de.txt",
    },
    "diffmark": {
        "name": "diffmark license",
        "url": "https://fedoraproject.org/wiki/Licensing/diffmark",
    },
    "WTFPL": {
        "name": "Do What The F*ck You Want To Public License",
        "url": "http://sam.zoy.org/wtfpl/COPYING",
    },
    "DOC": {
        "name": "DOC License",
        "url": "http://www.cs.wustl.edu/~schmidt/ACE-copying.html",
    },
    "Dotseqn": {
        "name": "Dotseqn License",
        "url": "https://fedoraproject.org/wiki/Licensing/Dotseqn",
    },
    "DSDP": {
        "name": "DSDP License",
        "url": "https://fedoraproject.org/wiki/Licensing/DSDP",
    },
    "dvipdfm": {
        "name": "dvipdfm License",
        "url": "https://fedoraproject.org/wiki/Licensing/dvipdfm",
    },
    "EPL-1.0": {
        "name": "Eclipse Public License 1.0",
        "url": "http://www.opensource.org/licenses/EPL-1.0",
    },
    "ECL-1.0": {
        "name": "Educational Community License v1.0",
        "url": "http://opensource.org/licenses/ECL-1.0",
    },
    "ECL-2.0": {
        "name": "Educational Community License v2.0",
        "url": "http://opensource.org/licenses/ECL-2.0",
    },
    "EFL-1.0": {
        "name": "Eiffel Forum License v1.0",
        "url": "http://opensource.org/licenses/EFL-1.0",
    },
    "EFL-2.0": {
        "name": "Eiffel Forum License v2.0",
        "url": "http://opensource.org/licenses/EFL-2.0",
    },
    "MIT-advertising": {
        "name": "Enlightenment License (e16)",
        "url": "https://fedoraproject.org/wiki/Licensing/MIT_With_Advertising",
    },
    "MIT-enna": {
        "name": "enna License",
        "url": "https://fedoraproject.org/wiki/Licensing/MIT#enna",
    },
    "Entessa": {
        "name": "Entessa Public License v1.0",
        "url": "http://opensource.org/licenses/Entessa",
    },
    "ErlPL-1.1": {
        "name": "Erlang Public License v1.1",
        "url": "http://www.erlang.org/EPLICENSE",
    },
    "EUDatagrid": {
        "name": "EU DataGrid Software License",
        "url": "http://www.opensource.org/licenses/EUDatagrid",
    },
    "EUPL-1.0": {
        "name": "European Union Public License 1.0",
        "url": "http://ec.europa.eu/idabc/en/document/7330.html",
    },
    "EUPL-1.1": {
        "name": "European Union Public License 1.1",
        "url": "http://www.opensource.org/licenses/EUPL-1.1",
    },
    "Eurosym": {
        "name": "Eurosym License",
        "url": "https://fedoraproject.org/wiki/Licensing/Eurosym",
    },
    "Fair": {
        "name": "Fair License",
        "url": "http://www.opensource.org/licenses/Fair",
    },
    "MIT-feh": {
        "name": "feh License",
        "url": "https://fedoraproject.org/wiki/Licensing/MIT#feh",
    },
    "Frameworx-1.0": {
        "name": "Frameworx Open License 1.0",
        "url": "http://www.opensource.org/licenses/Frameworx-1.0",
    },
    "FreeImage": {
        "name": "FreeImage Public License v1.0",
        "url": "http://freeimage.sourceforge.net/freeimage-license.txt",
    },
    "FTL": {
        "name": "Freetype Project License",
        "url": "http://freetype.fis.uniroma2.it/FTL.TXT",
    },
    "FSFUL": {
        "name": "FSF Unlimited License",
        "url": "https://fedoraproject.org/wiki/Licensing/FSF_Unlimited_License",
    },
    "FSFULLR": {
        "name": "FSF Unlimited License (with License Retention)",
        "url": "https://fedoraproject.org/wiki/Licensing/FSF_Unlimited_License",
    },
    "Giftware": {
        "name": "Giftware License",
        "url": "http://alleg.sourceforge.net//license.html",
    },
    "GL2PS": {
        "name": "GL2PS License",
        "url": "http://www.geuz.org/gl2ps/COPYING.GL2PS",
    },
    "Glulxe": {
        "name": "Glulxe License",
        "url": "https://fedoraproject.org/wiki/Licensing/Glulxe",
    },
    "AGPL-3.0": {
        "name": "GNU Affero General Public License v3.0",
        "url": "http://www.gnu.org/licenses/agpl.txt",
    },
    "GFDL-1.1": {
        "name": "GNU Free Documentation License v1.1",
        "url": "http://www.gnu.org/licenses/old-licenses/fdl-1.1.txt",
    },
    "GFDL-1.2": {
        "name": "GNU Free Documentation License v1.2",
        "url": "http://www.gnu.org/licenses/old-licenses/fdl-1.2.txt",
    },
    "GFDL-1.3": {
        "name": "GNU Free Documentation License v1.3",
        "url": "http://www.gnu.org/licenses/fdl-1.3.txt",
    },
    "GPL-1.0": {
        "name": "GNU General Public License v1.0 only",
        "url": "http://www.gnu.org/licenses/old-licenses/gpl-1.0-standalone.html",
    },
    "GPL-2.0": {
        "name": "GNU General Public License v2.0 only",
        "url": "http://www.opensource.org/licenses/GPL-2.0",
    },
    "GPL-3.0": {
        "name": "GNU General Public License v3.0 only",
        "url": "http://www.opensource.org/licenses/GPL-3.0",
    },
    "LGPL-2.1": {
        "name": "GNU Lesser General Public License v2.1 only",
        "url": "http://www.opensource.org/licenses/LGPL-2.1",
    },
    "LGPL-3.0": {
        "name": "GNU Lesser General Public License v3.0 only",
        "url": "http://www.opensource.org/licenses/LGPL-3.0",
    },
    "LGPL-2.0": {
        "name": "GNU Library General Public License v2 only",
        "url": "http://www.gnu.org/licenses/old-licenses/lgpl-2.0-standalone.html",
    },
    "gnuplot": {
        "name": "gnuplot License",
        "url": "https://fedoraproject.org/wiki/Licensing/Gnuplot",
    },
    "gSOAP-1.3b": {
        "name": "gSOAP Public License v1.3b",
        "url": "http://www.cs.fsu.edu/~engelen/license.html",
    },
    "HaskellReport": {
        "name": "Haskell Language Report License",
        "url": "https://fedoraproject.org/wiki/Licensing/Haskell_Language_Report_License",
    },
    "HPND": {
        "name": "Historic Permission Notice and Disclaimer",
        "url": "http://www.opensource.org/licenses/HPND",
    },
    "IPL-1.0": {
        "name": "IBM Public License v1.0",
        "url": "http://www.opensource.org/licenses/IPL-1.0",
    },
    "ICU": {
        "name": "ICU License",
        "url": "http://source.icu-project.org/repos/icu/icu/trunk/license.html",
    },
    "ImageMagick": {
        "name": "ImageMagick License",
        "url": "http://www.imagemagick.org/script/license.php",
    },
    "iMatix": {
        "name": "iMatix Standard Function Library Agreement",
        "url": "http://legacy.imatix.com/html/sfl/sfl4.htm#license",
    },
    "Imlib2": {
        "name": "Imlib2 License",
        "url": "http://trac.enlightenment.org/e/browser/trunk/imlib2/COPYING",
    },
    "IJG": {
        "name": "Independent JPEG Group License",
        "url": "http://dev.w3.org/cvsweb/Amaya/libjpeg/Attic/README?rev=1.2",
    },
    "Intel": {
        "name": "Intel Open Source License",
        "url": "http://opensource.org/licenses/Intel",
    },
    "IPA": {
        "name": "IPA Font License",
        "url": "http://www.opensource.org/licenses/IPA",
    },
    "JasPer-2.0": {
        "name": "JasPer License",
        "url": "http://www.ece.uvic.ca/~mdadams/jasper/LICENSE",
    },
    "JSON": {
        "name": "JSON License",
        "url": "http://www.json.org/license.html",
    },
    "LPPL-1.3a": {
        "name": "LaTeX Project Public License 1.3a",
        "url": "http://www.latex-project.org/lppl/lppl-1-3a.txt",
    },
    "LPPL-1.0": {
        "name": "LaTeX Project Public License v1.0",
        "url": "http://www.latex-project.org/lppl/lppl-1-0.txt",
    },
    "LPPL-1.1": {
        "name": "LaTeX Project Public License v1.1",
        "url": "http://www.latex-project.org/lppl/lppl-1-1.txt",
    },
    "LPPL-1.2": {
        "name": "LaTeX Project Public License v1.2",
        "url": "http://www.latex-project.org/lppl/lppl-1-2.txt",
    },
    "LPPL-1.3c": {
        "name": "LaTeX Project Public License v1.3c",
        "url": "http://www.opensource.org/licenses/LPPL-1.3c",
    },
    "Latex2e": {
        "name": "Latex2e License",
        "url": "https://fedoraproject.org/wiki/Licensing/Latex2e",
    },
    "BSD-3-Clause-LBNL": {
        "name": "Lawrence Berkeley National Labs BSD variant license",
        "url": "https://fedoraproject.org/wiki/Licensing/LBNLBSD",
    },
    "Leptonica": {
        "name": "Leptonica License",
        "url": "https://fedoraproject.org/wiki/Licensing/Leptonica",
    },
    "LGPLLR": {
        "name": "Lesser General Public License For Linguistic Resources",
        "url": "http://www-igm.univ-mlv.fr/~unitex/lgpllr.html",
    },
    "Libpng": {
        "name": "libpng License",
        "url": "http://www.libpng.org/pub/png/src/libpng-LICENSE.txt",
    },
    "libtiff": {
        "name": "libtiff License",
        "url": "https://fedoraproject.org/wiki/Licensing/libtiff",
    },
    "LPL-1.02": {
        "name": "Lucent Public License v1.02",
        "url": "http://www.opensource.org/licenses/LPL-1.02",
    },
    "LPL-1.0": {
        "name": "Lucent Public License Version 1.0",
        "url": "http://opensource.org/licenses/LPL-1.0",
    },
    "MakeIndex": {
        "name": "MakeIndex License",
        "url": "https://fedoraproject.org/wiki/Licensing/MakeIndex",
    },
    "MTLL": {
        "name": "Matrix Template Library License",
        "url": "https://fedoraproject.org/wiki/Licensing/Matrix_Template_Library_License",
    },
    "MS-PL": {
        "name": "Microsoft Public License",
        "url": "http://www.opensource.org/licenses/MS-PL",
    },
    "MS-RL": {
        "name": "Microsoft Reciprocal License",
        "url": "http://www.opensource.org/licenses/MS-RL",
    },
    "MirOS": {
        "name": "MirOS Licence",
        "url": "http://www.opensource.org/licenses/MirOS",
    },
    "MITNFA": {
        "name": "MIT +no-false-attribs license",
        "url": "https://fedoraproject.org/wiki/Licensing/MITNFA",
    },
    "MIT": {
        "name": "MIT License",
        "url": "http://www.opensource.org/licenses/MIT",
    },
    "Motosoto": {
        "name": "Motosoto License",
        "url": "http://www.opensource.org/licenses/Motosoto",
    },
    "MPL-1.0": {
        "name": "Mozilla Public License 1.0",
        "url": "http://www.mozilla.org/MPL/MPL-1.0.html",
    },
    "MPL-1.1": {
        "name": "Mozilla Public License 1.1",
        "url": "http://www.mozilla.org/MPL/MPL-1.1.html",
    },
    "MPL-2.0": {
        "name": "Mozilla Public License 2.0",
        "url": "http://www.mozilla.org/MPL/2.0/\nhttp://opensource.org/licenses/MPL-2.0",
    },
    "MPL-2.0-no-copyleft-exception": {
        "name": "Mozilla Public License 2.0 (no copyleft exception)",
        "url": "http://www.mozilla.org/MPL/2.0/\nhttp://opensource.org/licenses/MPL-2.0",
    },
    "mpich2": {
        "name": "mpich2 License",
        "url": "https://fedoraproject.org/wiki/Licensing/MIT",
    },
    "Multics": {
        "name": "Multics License",
        "url": "http://www.opensource.org/licenses/Multics",
    },
    "Mup": {
        "name": "Mup License",
        "url": "https://fedoraproject.org/wiki/Licensing/Mup",
    },
    "NASA-1.3": {
        "name": "NASA Open Source Agreement 1.3",
        "url": "http://www.opensource.org/licenses/NASA-1.3",
    },
    "Naumen": {
        "name": "Naumen Public License",
        "url": "http://www.opensource.org/licenses/Naumen",
    },
    "NetCDF": {
        "name": "NetCDF license",
        "url": "http://www.unidata.ucar.edu/software/netcdf/copyright.html",
    },
    "NGPL": {
        "name": "Nethack General Public License",
        "url": "http://www.opensource.org/licenses/NGPL",
    },
    "NOSL": {
        "name": "Netizen Open Source License",
        "url": "http://bits.netizen.com.au/licenses/NOSL/nosl.txt",
    },
    "NPL-1.0": {
        "name": "Netscape Public License v1.0",
        "url": "http://www.mozilla.org/MPL/NPL/1.0/",
    },
    "NPL-1.1": {
        "name": "Netscape Public License v1.1",
        "url": "http://www.mozilla.org/MPL/NPL/1.1/",
    },
    "Newsletr": {
        "name": "Newsletr License",
        "url": "https://fedoraproject.org/wiki/Licensing/Newsletr",
    },
    "NLPL": {
        "name": "No Limit Public License",
        "url": "https://fedoraproject.org/wiki/Licensing/NLPL",
    },
    "Nokia": {
        "name": "Nokia Open Source License",
        "url": "http://www.opensource.org/licenses/nokia",
    },
    "NPOSL-3.0": {
        "name": "Non-Profit Open Software License 3.0",
        "url": "http://www.opensource.org/licenses/NOSL3.0",
    },
    "Noweb": {
        "name": "Noweb License",
        "url": "https://fedoraproject.org/wiki/Licensing/Noweb",
    },
    "NRL": {
        "name": "NRL License",
        "url": "http://web.mit.edu/network/isakmp/nrllicense.html",
    },
    "NTP": {
        "name": "NTP License",
        "url": "http://www.opensource.org/licenses/NTP",
    },
    "Nunit": {
        "name": "Nunit License",
        "url": "https://fedoraproject.org/wiki/Licensing/Nunit",
    },
    "OCLC-2.0": {
        "name": "OCLC Research Public License 2.0",
        "url": "http://www.opensource.org/licenses/OCLC-2.0",
    },
    "ODbL-1.0": {
        "name": "ODC Open Database License v1.0",
        "url": "http://www.opendatacommons.org/licenses/odbl/1.0/",
    },
    "PDDL-1.0": {
        "name": "ODC Public Domain Dedication & License 1.0",
        "url": "http://opendatacommons.org/licenses/pddl/1.0/",
    },
    "OGTSL": {
        "name": "Open Group Test Suite License",
        "url": "http://www.opensource.org/licenses/OGTSL",
    },
    "OML": {
        "name": "Open Market License",
        "url": "https://fedoraproject.org/wiki/Licensing/Open_Market_License",
    },
    "OPL-1.0": {
        "name": "Open Public License v1.0",
        "url": "https://fedoraproject.org/wiki/Licensing/Open_Public_License",
    },
    "OSL-1.0": {
        "name": "Open Software License 1.0",
        "url": "http://opensource.org/licenses/OSL-1.0",
    },
    "OSL-1.1": {
        "name": "Open Software License 1.1",
        "url": "https://fedoraproject.org/wiki/Licensing/OSL1.1",
    },
    "PHP-3.01": {
        "name": "PHP License v3.01",
        "url": "http://www.php.net/license/3_01.txt",
    },
    "Plexus": {
        "name": "Plexus Classworlds License",
        "url": "https://fedoraproject.org/wiki/Licensing/Plexus_Classworlds_License",
    },
    "PostgreSQL": {
        "name": "PostgreSQL License",
        "url": "http://www.opensource.org/licenses/PostgreSQL",
    },
    "psfrag": {
        "name": "psfrag License",
        "url": "https://fedoraproject.org/wiki/Licensing/psfrag",
    },
    "psutils": {
        "name": "psutils License",
        "url": "https://fedoraproject.org/wiki/Licensing/psutils",
    },
    "Python-2.0": {
        "name": "Python License 2.0",
        "url": "http://www.opensource.org/licenses/Python-2.0",
    },
    "QPL-1.0": {
        "name": "Q Public License 1.0",
        "url": "http://www.opensource.org/licenses/QPL-1.0",
    },
    "Qhull": {
        "name": "Qhull License",
        "url": "https://fedoraproject.org/wiki/Licensing/Qhull",
    },
    "Rdisc": {
        "name": "Rdisc License",
        "url": "https://fedoraproject.org/wiki/Licensing/Rdisc_License",
    },
    "RPSL-1.0": {
        "name": "RealNetworks Public Source License v1.0",
        "url": "http://www.opensource.org/licenses/RPSL-1.0",
    },
    "RPL-1.1": {
        "name": "Reciprocal Public License 1.1",
        "url": "http://opensource.org/licenses/RPL-1.1",
    },
    "RPL-1.5": {
        "name": "Reciprocal Public License 1.5",
        "url": "http://www.opensource.org/licenses/RPL-1.5",
    },
    "RHeCos-1.1": {
        "name": "Red Hat eCos Public License v1.1",
        "url": "http://ecos.sourceware.org/old-license.html",
    },
    "RSCPL": {
        "name": "Ricoh Source Code Public License",
        "url": "http://www.opensource.org/licenses/RSCPL",
    },
    "RSA-MD": {
        "name": "RSA Message-Digest License",
        "url": "http://www.faqs.org/rfcs/rfc1321.html",
    },
    "Ruby": {
        "name": "Ruby License",
        "url": "http://www.ruby-lang.org/en/LICENSE.txt",
    },
    "SAX-PD": {
        "name": "Sax Public Domain Notice",
        "url": "http://www.saxproject.org/copying.html",
    },
    "Saxpath": {
        "name": "Saxpath License",
        "url": "https://fedoraproject.org/wiki/Licensing/Saxpath_License",
    },
    "SCEA": {
        "name": "SCEA Shared Source License",
        "url": "http://research.scea.com/scea_shared_source_license.html",
    },
    "SWL": {
        "name": "Scheme Widget Library (SWL) Software License Agreement",
        "url": "https://fedoraproject.org/wiki/Licensing/SWL",
    },
    "Sendmail": {
        "name": "Sendmail License",
        "url": "http://www.sendmail.com/pdfs/open_source/sendmail_license.pdf",
    },
    "SGI-B-1.0": {
        "name": "SGI Free Software License B v1.0",
        "url": "http://oss.sgi.com/projects/FreeB/SGIFreeSWLicB.1.0.html",
    },
    "SGI-B-1.1": {
        "name": "SGI Free Software License B v1.1",
        "url": "http://oss.sgi.com/projects/FreeB/",
    },
    "SGI-B-2.0": {
        "name": "SGI Free Software License B v2.0",
        "url": "http://oss.sgi.com/projects/FreeB/SGIFreeSWLicB.2.0.pdf",
    },
    "OFL-1.0": {
        "name": "SIL Open Font License 1.0",
        "url": "http://scripts.sil.org/cms/scripts/page.php?item_id=OFL10_web",
    },
    "OFL-1.1": {
        "name": "SIL Open Font License 1.1",
        "url": "http://www.opensource.org/licenses/OFL-1.1",
    },
    "SimPL-2.0": {
        "name": "Simple Public License 2.0",
        "url": "http://www.opensource.org/licenses/SimPL-2.0",
    },
    "Sleepycat": {
        "name": "Sleepycat License",
        "url": "http://www.opensource.org/licenses/Sleepycat",
    },
    "SNIA": {
        "name": "SNIA Public License 1.1",
        "url": "https://fedoraproject.org/wiki/Licensing/SNIA_Public_License",
    },
    "SMLNJ": {
        "name": "Standard ML of New Jersey License",
        "url": "http://www.smlnj.org//license.html",
    },
    "SugarCRM-1.1.3": {
        "name": "SugarCRM Public License v1.1.3",
        "url": "http://www.sugarcrm.com/crm/SPL",
    },
    "SISSL": {
        "name": "Sun Industry Standards Source License v1.1",
        "url": "http://opensource.org/licenses/SISSL",
    },
    "SISSL-1.2": {
        "name": "Sun Industry Standards Source License v1.2",
        "url": "http://gridscheduler.sourceforge.net/Gridengine_SISSL_license.html",
    },
    "SPL-1.0": {
        "name": "Sun Public License v1.0",
        "url": "http://www.opensource.org/licenses/SPL-1.0",
    },
    "Watcom-1.0": {
        "name": "Sybase Open Watcom Public License 1.0",
        "url": "http://www.opensource.org/licenses/Watcom-1.0",
    },
    "TCL": {
        "name": "TCL/TK License",
        "url": "https://fedoraproject.org/wiki/Licensing/TCL",
    },
    "Unlicense": {
        "name": "The Unlicense",
        "url": "http://unlicense.org/",
    },
    "TMate": {
        "name": "TMate Open Source License",
        "url": "http://svnkit.com/license.html",
    },
    "TORQUE-1.1": {
        "name": "TORQUE v2.5+ Software License v1.1",
        "url": "https://fedoraproject.org/wiki/Licensing/TORQUEv1.1",
    },
    "TOSL": {
        "name": "Trusster Open Source License",
        "url": "https://fedoraproject.org/wiki/Licensing/TOSL",
    },
    "Unicode-TOU": {
        "name": "Unicode Terms of Use",
        "url": "http://www.unicode.org/copyright.html",
    },
    "UPL-1.0": {
        "name": "Universal Permissive License v1.0",
        "url": "http://opensource.org/licenses/UPL",
    },
    "NCSA": {
        "name": "University of Illinois/NCSA Open Source License",
        "url": "http://www.opensource.org/licenses/NCSA",
    },
    "Vim": {
        "name": "Vim License",
        "url": "http://vimdoc.sourceforge.net/htmldoc/uganda.html",
    },
    "VOSTROM": {
        "name": "VOSTROM Public License for Open Source",
        "url": "https://fedoraproject.org/wiki/Licensing/VOSTROM",
    },
    "VSL-1.0": {
        "name": "Vovida Software License v1.0",
        "url": "http://www.opensource.org/licenses/VSL-1.0",
    },
    "W3C-19980720": {
        "name": "W3C Software Notice and License (1998-07-20)",
        "url": "http://www.w3.org/Consortium/Legal/copyright-software-19980720.html",
    },
    "W3C": {
        "name": "W3C Software Notice and License (2002-12-31)",
        "url": "http://www.opensource.org/licenses/W3C",
    },
    "Wsuipa": {
        "name": "Wsuipa License",
        "url": "https://fedoraproject.org/wiki/Licensing/Wsuipa",
    },
    "Xnet": {
        "name": "X.Net License",
        "url": "http://opensource.org/licenses/Xnet",
    },
    "X11": {
        "name": "X11 License",
        "url": "http://www.xfree86.org/3.3.6/COPYRIGHT2.html#3",
    },
    "Xerox": {
        "name": "Xerox License",
        "url": "https://fedoraproject.org/wiki/Licensing/Xerox",
    },
    "XFree86-1.1": {
        "name": "XFree86 License 1.1",
        "url": "http://www.xfree86.org/current/LICENSE4.html",
    },
    "xinetd": {
        "name": "xinetd License",
        "url": "https://fedoraproject.org/wiki/Licensing/Xinetd_License",
    },
    "xpp": {
        "name": "XPP License",
        "url": "https://fedoraproject.org/wiki/Licensing/xpp",
    },
    "XSkat": {
        "name": "XSkat License",
        "url": "https://fedoraproject.org/wiki/Licensing/XSkat_License",
    },
    "YPL-1.0": {
        "name": "Yahoo! Public License v1.0",
        "url": "http://www.zimbra.com/license/yahoo_public_license_1.0.html",
    },
    "YPL-1.1": {
        "name": "Yahoo! Public License v1.1",
        "url": "http://www.zimbra.com/license/yahoo_public_license_1.1.html",
    },
    "Zed": {
        "name": "Zed License",
        "url": "https://fedoraproject.org/wiki/Licensing/Zed",
    },
    "Zlib": {
        "name": "zlib License",
        "url": "http://www.opensource.org/licenses/Zlib",
    },
    "zlib-acknowledgement": {
        "name": "zlib/libpng License with Acknowledgement",
        "url": "https://fedoraproject.org/wiki/Licensing/ZlibWithAcknowledgement",
    },
    "ZPL-1.1": {
        "name": "Zope Public License 1.1",
        "url": "http://old.zope.org/Resources/License/ZPL-1.1",
    },
    "ZPL-2.0": {
        "name": "Zope Public License 2.0",
        "url": "http://opensource.org/licenses/ZPL-2.0",
    },
    "ZPL-2.1": {
        "name": "Zope Public License 2.1",
        "url": "http://old.zope.org/Resources/ZPL/",
    }
}


@attr.s
class License(object):
    id = attr.ib()
    name = attr.ib()
    url = attr.ib()

    @property
    def legalcode(self):
        p = pathlib.Path(__file__).parent / 'legalcode' / self.id
        if p.exists():
            return p.read_text(encoding='utf8')


_LICENSES = [License(id_, l['name'], l['url']) for id_, l in _LICENSES.items()]


def find(q):
    for license_ in _LICENSES:
        if q.lower() == license_.id.lower() or q == license_.name or q == license_.url:
            return license_
        if '://' in q:
            u1 = license_.url.split('://')[1]
            u2 = q.split('://')[1]
            if u1.startswith(u2) or u2.startswith(u1):
                return license_
